# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2018-2023, Intel Corporation

import argparse
import statemachine
import opscontext
import consistencycheckwrap
import loggingfacility
import markerparser
import sys
import reorderengines
import signal

# Prevent printing python stacktrace on SIGINT. To achieve that throw
# "SystemExit exception" instead of "KeyboardInterrupt exception".
signal.signal(signal.SIGINT, lambda signum, frame: exit(signum))


def main():
    pmreorder_version = "unknown"

    """
    Argv[1] should be given in order to use -v or --version flag. It is passed
    from the installation script. We check whether argv[1] was given. If it's
    not any of regular parameters - we use it as a version of pmreorder and
    remove it from the arguments list.
    """
    if len(sys.argv) > 1 and sys.argv[1][0] != "-":
        pmreorder_version = sys.argv[1]
        del sys.argv[1]

    # TODO unicode support
    # TODO parameterize reorder engine type
    parser = argparse.ArgumentParser(description="Store reordering tool")
    parser.add_argument(
        "-l",
        "--logfile",
        required=True,
        help="the pmemcheck log file to process",
    )
    parser.add_argument(
        "-c",
        "--checker",
        choices=consistencycheckwrap.checkers,
        default=consistencycheckwrap.checkers[0],
        help="choose consistency checker type",
    )
    parser.add_argument(
        "-p",
        "--path",
        required=True,
        help="path to the consistency checker and arguments. "
        + "Note: If program is given, the program has to take "
        + "a file name as the last parameter.",
        nargs="+",
    )
    parser.add_argument(
        "-n",
        "--name",
        help="consistency check function for the 'lib' checker. Note: "
        + "The function has to take a file name as the only parameter.",
    )
    parser.add_argument("-o", "--output", help="set the logger output file")
    parser.add_argument(
        "-e",
        "--output-level",
        choices=loggingfacility.log_levels,
        help="set the output log level",
    )
    parser.add_argument(
        "-x",
        "--extended-macros",
        help="list of pairs MARKER=ENGINE or json config file",
    )
    parser.add_argument(
        "-v",
        "--version",
        help="print version of the pmreorder",
        action="version",
        version="%(prog)s " + pmreorder_version,
    )
    engines_keys = list(reorderengines.engines.keys())
    parser.add_argument(
        "-r",
        "--default-engine",
        help="set default reorder engine default=NoReorderNoChecker",
        choices=engines_keys,
        default=engines_keys[0],
    )
    args = parser.parse_args()
    logger = loggingfacility.get_logger(args.output, args.output_level)
    checker = consistencycheckwrap.get_checker(
        args.checker, " ".join(args.path), args.name, logger
    )

    markers = markerparser.MarkerParser().get_markers(args.extended_macros)

    # create the script context
    context = opscontext.OpsContext(
        args.logfile, checker, logger, args.default_engine, markers
    )
    logger.debug("Input parameters: {}".format(context.__dict__))

    # init and run the state machine
    a = statemachine.StateMachine(statemachine.InitState(context))
    operations, operation_ids, markers = context.extract_operations()
    if a.run_all(operations, operation_ids, markers) is False:
        sys.exit(1)


if __name__ == "__main__":
    main()



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation
#
import testframework as t
from testframework import granularity as g


@g.no_testdir()
class PMEM2_DEEP_FLUSH(t.Test):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('pmem2_deep_flush', self.test_case)


class TEST0(PMEM2_DEEP_FLUSH):
    """test pmem2_deep_flush"""
    test_case = "test_deep_flush_func"


class TEST1(PMEM2_DEEP_FLUSH):
    """test pmem2_deep_flush with mocked DAX devices"""
    test_case = "test_deep_flush_func_devdax"


class TEST2(PMEM2_DEEP_FLUSH):
    """test pmem2_deep_flush with range beyond mapping"""
    test_case = "test_deep_flush_range_beyond_mapping"



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2021, Intel Corporation
#


import os

import testframework as t
import futils


@t.require_valgrind_enabled('pmemcheck')
class Pmem2ApiLogs(t.Test):
    test_type = t.Medium

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile')
        ctx.env['PMREORDER_EMIT_LOG'] = self.PMREORDER_EMIT_LOG
        ctx.valgrind.add_opt('--log-stores=yes')
        ctx.exec('pmem2_api', self.test_case, filepath)

        log_name = 'pmemcheck{}.log'.format(self.testnum)
        pmemecheck_log = os.path.join(ctx.cwd, log_name)
        memmove_fn_begin_nums = futils.count(
            pmemecheck_log, 'pmem2_memmove.BEGIN')
        memmove_fn_end_nums = futils.count(
            pmemecheck_log, 'pmem2_memmove.END')
        memset_fn_begin_nums = futils.count(
            pmemecheck_log, 'pmem2_memset.BEGIN')
        memset_fn_end_nums = futils.count(
            pmemecheck_log, 'pmem2_memset.END')

        if (memmove_fn_begin_nums != self.expected_memmove_fn_nums or
                memmove_fn_end_nums != self.expected_memmove_fn_nums or
                memset_fn_begin_nums != self.expected_memset_fn_nums or
                memset_fn_end_nums != self.expected_memset_fn_nums):
            raise futils.Fail(
                'Pattern: pmem2_memmove.BEGIN occurs {} times. Expected {}.\n'
                'Pattern: pmem2_memmove.END occurs {} times. Expected {}.\n'
                'Pattern: pmem2_memset.BEGIN occurs {} times. Expected {}.\n'
                'Pattern: pmem2_memset.END occurs {} times. Expected {}.'
                .format(memmove_fn_begin_nums, self.expected_memmove_fn_nums,
                        memmove_fn_end_nums, self.expected_memmove_fn_nums,
                        memset_fn_begin_nums, self.expected_memset_fn_nums,
                        memset_fn_end_nums, self.expected_memset_fn_nums)
            )


class TEST0(Pmem2ApiLogs):
    """
    test the emission of library and function names to pmemcheck stores log
    """
    test_case = "test_pmem2_api_logs"
    expected_memmove_fn_nums = 2
    expected_memset_fn_nums = 1
    PMREORDER_EMIT_LOG = '1'


class TEST1(Pmem2ApiLogs):
    """
    test the emission of library and function names to pmemcheck stores log
    with PMREORDER_EMIT_LOG set to '0'. Any names shouldn't be emitted.
    """
    test_case = "test_pmem2_api_logs"
    expected_memmove_fn_nums = 0
    expected_memset_fn_nums = 0
    PMREORDER_EMIT_LOG = '0'



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#


import os
from enum import Enum

import testframework as t
from testframework import granularity as g


class Granularity(Enum):
    BYTE = 1
    CACHE_LINE = 2
    PAGE = 3


@g.require_granularity(g.ANY)
class PMEM2_GRANULARITY(t.BaseTest):
    test_type = t.Short
    available_granularity = None

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile1')

        bus_dev_path_without_eADR = os.path.join(
            self.cwd,
            "linux_eadr_paths/eadr_not_available/sys/bus/nd/devices/")
        bus_dev_path_with_eADR = os.path.join(
            self.cwd,
            "linux_eadr_paths/eadr_available/sys/bus/nd/devices/")

        # Testframework may set this variable to emulate the certain type of
        # granularity.
        # This test mocks all granularity checks but they are skipped if
        # granularity is forced so this test requires unforced granularity.
        ctx.env['PMEM2_FORCE_GRANULARITY'] = '0'

        if self.available_granularity == Granularity.BYTE:
            ctx.env['IS_EADR'] = '1'
            ctx.env['IS_PMEM'] = '1'
            ctx.env['BUS_DEVICE_PATH'] = bus_dev_path_with_eADR
        elif self.available_granularity == Granularity.CACHE_LINE:
            ctx.env['IS_EADR'] = '0'
            ctx.env['IS_PMEM'] = '1'
            ctx.env['BUS_DEVICE_PATH'] = bus_dev_path_without_eADR
        elif self.available_granularity == Granularity.PAGE:
            ctx.env['IS_EADR'] = '0'
            ctx.env['IS_PMEM'] = '0'
            ctx.env['BUS_DEVICE_PATH'] = bus_dev_path_without_eADR

        ctx.exec('pmem2_granularity', self.test_case, filepath)


class TEST0(PMEM2_GRANULARITY):
    """pass byte granularity, available byte granularity"""
    test_case = "test_granularity_req_byte_avail_byte"
    available_granularity = Granularity.BYTE


class TEST2(PMEM2_GRANULARITY):
    """pass byte granularity, available cache line granularity"""
    test_case = "test_granularity_req_byte_avail_cl"
    available_granularity = Granularity.CACHE_LINE


class TEST4(PMEM2_GRANULARITY):
    """pass byte granularity, available page granularity"""
    test_case = "test_granularity_req_byte_avail_page"
    available_granularity = Granularity.PAGE


class TEST6(PMEM2_GRANULARITY):
    """pass cache line granularity, available byte granularity"""
    test_case = "test_granularity_req_cl_avail_byte"
    available_granularity = Granularity.BYTE


class TEST7(PMEM2_GRANULARITY):
    """pass cache line granularity, available cache line granularity"""
    test_case = "test_granularity_req_cl_avail_cl"
    available_granularity = Granularity.CACHE_LINE


class TEST9(PMEM2_GRANULARITY):
    """pass cache line granularity, available page granularity"""
    test_case = "test_granularity_req_cl_avail_page"
    available_granularity = Granularity.PAGE


class TEST11(PMEM2_GRANULARITY):
    """pass page granularity, available byte granularity"""
    test_case = "test_granularity_req_page_avail_byte"
    available_granularity = Granularity.BYTE


class TEST12(PMEM2_GRANULARITY):
    """pass page granularity, available cache line granularity"""
    test_case = "test_granularity_req_page_avail_cl"
    available_granularity = Granularity.CACHE_LINE


class TEST13(PMEM2_GRANULARITY):
    """pass page granularity, available page granularity"""
    test_case = "test_granularity_req_page_avail_page"
    available_granularity = Granularity.PAGE



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2023, Intel Corporation

import argparse
import json

from typing import List, Dict, Any

# https://peps.python.org/pep-0589/ requires Python >= 3.8
# from typing import TypedDict

# List all API calls which start call stacks containing a particular function name.

PARSER = argparse.ArgumentParser()
PARSER.add_argument('-a', '--call-stacks-all', default='call_stacks_all.json')
PARSER.add_argument('-f', '--function-name', required=True)

# class CallStack(TypedDict): # for Python >= 3.8
#     stack: list[str]
#     size: int
CallStack = Dict[str, Any] # for Python < 3.8

def load_call_stacks_from_json(file_name: str) -> List[CallStack]:
        with open(file_name, 'r') as file:
                return json.load(file)

def main():
        args = PARSER.parse_args()
        call_stacks = load_call_stacks_from_json(args.call_stacks_all)
        apis = []
        # lookup all call stacks in which the function of interest is mentioned
        for call_stack in call_stacks:
                if args.function_name in call_stack['stack']:
                        # callect all API calls which starts these call stacks
                        apis.append(call_stack['stack'][0])
        # remove duplicates
        apis = list(set(apis))
        apis.sort()
        for api in apis:
                print(api)

if __name__ == '__main__':
        main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2018-2022, Intel Corporation

from operationfactory import OperationFactory
from binaryoutputhandler import BinaryOutputHandler
import reorderengines
import memoryoperations
from itertools import repeat


class OpsContext:
    """
    Holds the context of the performed operations.

    :ivar _operations: The operations to be performed, based on the log file.
    :type _operations: list of strings
    :ivar reorder_engine: The reordering engine used at the moment.
    :type one of the reorderengine Class
    :ivar default_engine: The default reordering engine.
    :type default_engine: One of the reorderengines Class
    :ivar test_on_barrier: Check consistency on barrier.
    :type test_on_barrier: bool
    :ivar default_barrier: Default consistency barrier status.
    :type default_barrier: bool
    :ivar file_handler: The file handler used.
    """

    def __init__(self, log_file, checker, logger, arg_engine, markers):
        """
        Splits the operations in the log file and sets the instance variables
        to default values.

        :param log_file: The full name of the log file.
        :type log_file: str
        :return: None
        """
        # TODO reading the whole file at once is rather naive
        # change in the future
        self._operations = open(log_file).read().split("|")
        engine = reorderengines.get_engine(arg_engine)
        self.reorder_engine = engine
        self.test_on_barrier = engine.test_on_barrier
        self.default_engine = self.reorder_engine
        self.default_barrier = self.default_engine.test_on_barrier
        self.file_handler = BinaryOutputHandler(checker, logger)
        self.checker = checker
        self.logger = logger
        self.markers = markers
        self.stack_engines = [("START", getattr(memoryoperations, arg_engine))]

    # TODO this should probably be made a generator
    def extract_operations(self):
        """
        Creates specific operation objects based on the labels available
        in the split log file.

        :return: list of subclasses of :class:`memoryoperations.BaseOperation`
        """
        enumerated_ops = list(enumerate(self._operations))
        markers = list(
            filter(
                lambda e: e[1].endswith(".BEGIN") or e[1].endswith(".END"),
                enumerated_ops,
            )
        )
        operation_ids = list(enumerated_ops)

        stop_index = start_index = 0

        for i, elem in enumerated_ops:
            if "START" in elem:
                start_index = i
            elif "STOP" in elem:
                stop_index = i

        operations = list(
            map(
                OperationFactory.create_operation,
                self._operations[start_index + 1: stop_index],
                repeat(self.markers),
                repeat(self.stack_engines),
            )
        )

        return (operations, operation_ids, markers)



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2023-2024, Intel Corporation

import json
from typing import Dict, List

# https://peps.python.org/pep-0613/ requires Python >= 3.10
# from typing import TypeAlias

# Calls: TypeAlias = dict[str, list[str]] # for Python >= 3.10
Calls = Dict[str, List[str]] # for Python < 3.10

def dict_extend(dict_, key, values):
        if key not in dict_.keys():
                dict_[key] = values
        else:
                dict_[key].extend(values)
        return dict_

def inlines(calls: Calls) -> Calls:
        # common
        calls['core_init'] = ['util_init', 'core_log_init', 'out_init']
        calls['core_fini'] = ['out_fini', 'core_log_fini', 'last_error_msg_fini']
        calls['common_init'] = ['core_init', 'util_mmap_init']
        calls['common_fini'] = ['util_mmap_fini', 'core_fini']
        calls['core_log_init'] = ['core_log_default_init', 'core_log_set_function']

        # libpmem
        calls['flush_empty'] = ['flush_empty_nolog']

        # libpmemobj
        calls['libpmemobj_init'] = ['common_init']
        calls['run_vg_init'] = ['run_iterate_used']
        calls['palloc_heap_action_on_unlock'] = ['palloc_reservation_clear']
        calls['palloc_heap_action_on_cancel'] = ['palloc_reservation_clear']
        calls['util_uuid_generate'] = ['util_uuid_from_string']

        return calls

def core_function_pointers(calls: Calls) -> Calls:
        calls['core_log_va'] = ['core_log_default_function']
        return calls

def pmem_function_pointers(calls: Calls) -> Calls:
        calls['pmem_drain'] = ['fence_empty', 'memory_barrier']

        flush_all = ['flush_empty', 'flush_clflush', 'flush_clflushopt', 'flush_clwb']
        calls['pmem_deep_flush'] = flush_all
        calls['pmem_flush'] = flush_all

        # '.static' suffix added to differentiate between libpmem API function and a static helper.
        memmove_nodrain_all = ['memmove_nodrain_libc', 'memmove_nodrain_generic', 'pmem2_memmove_nodrain', 'pmem2_memmove_nodrain_eadr']
        calls['pmem_memmove'] = memmove_nodrain_all
        calls['pmem_memcpy'] = memmove_nodrain_all
        calls['pmem_memmove_nodrain'] = memmove_nodrain_all
        calls['pmem_memcpy_nodrain'] = memmove_nodrain_all
        calls['pmem_memmove_persist'] = memmove_nodrain_all
        calls['pmem_memcpy_persist'] = memmove_nodrain_all

        memset_nodrain_all = ['memset_nodrain_libc', 'memset_nodrain_generic', 'pmem2_memset_nodrain', 'pmem2_memset_nodrain_eadr']
        calls['pmem_memset'] = memset_nodrain_all
        calls['pmem_memset_nodrain'] = memset_nodrain_all
        calls['pmem_memset_persist'] = memset_nodrain_all

        memmove_funcs = {
                't': {
                        func: [ f'memmove_mov_{trick}_{func}'
                                for trick in ['sse2', 'avx', 'avx512f']
                        ] for func in ['noflush', 'empty']
                },
                'nt': {
                        'empty': [ f'memmove_movnt_{trick}_empty_{drain}'
                                for trick in ['sse2', 'avx']
                                        for drain in ['wcbarrier', 'nobarrier']
                        ],
                        'flush': [ f'memmove_movnt_{trick}_{flush}_{drain}'
                                for trick in ['sse2', 'avx']
                                        for flush in ['clflush', 'clflushopt', 'clwb']
                                                for drain in ['wcbarrier', 'nobarrier']
                        ]
                }
        }
        memmove_funcs_extras = {
                't': {
                        'flush': [ f'memmove_mov_{trick}_{flush}'
                                for trick in ['sse2', 'avx', 'avx512f']
                                        for flush in ['clflush', 'clflushopt', 'clwb']
                        ]
                },
                'nt': {
                        'empty': [ f'memmove_movnt_{trick}_empty'
                                for trick in ['avx512f', 'movdir64b']
                        ],
                        'flush': [ f'memmove_movnt_{trick}_{flush}'
                                for trick in ['avx512f', 'movdir64b']
                                        for flush in ['clflush', 'clflushopt', 'clwb']
                        ]
                }
        }
        memmove_funcs['t']['flush'] = memmove_funcs_extras['t']['flush']
        memmove_funcs['nt']['empty'].extend(memmove_funcs_extras['nt']['empty'])
        memmove_funcs['nt']['flush'].extend(memmove_funcs_extras['nt']['flush'])

        calls['pmem2_memmove_nodrain'] = \
                            memmove_funcs['t']['noflush'] + \
                            memmove_funcs['nt']['flush'] + \
                            memmove_funcs['t']['flush']

        calls['pmem2_memmove_nodrain_eadr'] = \
                            memmove_funcs['t']['noflush'] + \
                            memmove_funcs['nt']['empty'] + \
                            memmove_funcs['t']['empty']

        memsetfuncs = {
                't': {
                        func: [ f'memset_mov_{trick}_{func}'
                                for trick in ['sse2', 'avx', 'avx512f']
                        ] for func in ['noflush', 'empty']
                },
                'nt': {
                        'empty': [ f'memset_movnt_{trick}_empty_{drain}'
                                for trick in ['sse2', 'avx']
                                        for drain in ['wcbarrier', 'nobarrier']
                        ],
                        'flush': [ f'memset_movnt_{trick}_{flush}_{drain}'
                                for trick in ['sse2', 'avx']
                                        for flush in ['clflush', 'clflushopt', 'clwb']
                                                for drain in ['wcbarrier', 'nobarrier']
                        ]
                }
        }
        memsetfuncs_extras = {
                't': {
                        'flush': [ f'memset_mov_{trick}_{flush}'
                                for trick in ['sse2', 'avx', 'avx512f']
                                        for flush in ['clflush', 'clflushopt', 'clwb']
                        ]
                },
                'nt': {
                        'empty': [ f'memset_movnt_{trick}_empty'
                                for trick in ['avx512f', 'movdir64b']
                        ],
                        'flush': [ f'memset_movnt_{trick}_{flush}'
                                for trick in ['avx512f', 'movdir64b']
                                        for flush in ['clflush', 'clflushopt', 'clwb']
                        ]
                }
        }
        memsetfuncs['t']['flush'] = memsetfuncs_extras['t']['flush']
        memsetfuncs['nt']['empty'].extend(memsetfuncs_extras['nt']['empty'])
        memsetfuncs['nt']['flush'].extend(memsetfuncs_extras['nt']['flush'])

        calls['pmem2_memset_nodrain'] = \
                            memsetfuncs['t']['noflush'] + \
                            memsetfuncs['nt']['flush'] + \
                            memsetfuncs['t']['flush']

        calls['pmem2_memset_nodrain_eadr'] = \
                            memsetfuncs['t']['noflush'] + \
                            memsetfuncs['nt']['empty'] + \
                            memsetfuncs['t']['empty']

        is_pmem_all = ['is_pmem_never', 'is_pmem_always', 'is_pmem_detect']

        calls = dict_extend(calls, 'pmem_is_pmem', is_pmem_all)

        return calls

def pmemobj_function_pointers(calls: Calls) -> Calls:
        # block_container_ops
        insert_all = ['container_ravl_insert_block', 'container_seglists_insert_block']
        get_rm_exact_all = ['container_ravl_get_rm_block_exact']
        get_rm_bestfit_all = ['container_ravl_get_rm_block_bestfit', 'container_seglists_get_rm_block_bestfit']
        is_empty_all = ['container_ravl_is_empty', 'container_seglists_is_empty']
        rm_all_all = ['container_ravl_rm_all', 'container_seglists_rm_all']
        destroy_all = ['container_ravl_destroy', 'container_seglists_destroy']

        calls = dict_extend(calls, 'bucket_insert_block', insert_all)

        calls = dict_extend(calls, 'bucket_remove_block', get_rm_exact_all)

        calls = dict_extend(calls, 'bucket_alloc_block', get_rm_bestfit_all)

        callers = [
                'bucket_attach_run',
                'bucket_detach_run'
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, rm_all_all)

        calls = dict_extend(calls, 'bucket_fini', destroy_all)

        compare_all = ['ravl_interval_compare']

        callers = [
                'ravl_emplace',
                'ravl_find'
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, compare_all)

        # memory_block_ops
        block_size_all = ['huge_block_size', 'run_block_size']
        prep_hdr_all = ['huge_prep_operation_hdr', 'run_prep_operation_hdr']
        get_lock_all = ['huge_get_lock', 'run_get_lock']
        get_state_all = ['huge_get_state', 'run_get_state']
        get_user_data_all = ['block_get_user_data']
        get_real_data_all = ['huge_get_real_data', 'run_get_real_data']
        get_user_size_all = ['block_get_user_size']
        get_real_size_all = ['block_get_real_size']
        write_header_all = ['block_write_header']
        invalidate_all = ['block_invalidate']
        ensure_header_type_all = ['huge_ensure_header_type', 'run_ensure_header_type']
        reinit_header_all = ['block_reinit_header']
        vg_init_all = ['huge_vg_init', 'run_vg_init']
        get_extra_all = ['block_get_extra']
        get_flags_all = ['block_get_flags']
        iterate_free_all = ['huge_iterate_free', 'run_iterate_free']
        iterate_used_all = ['huge_iterate_used', 'run_iterate_used']
        reinit_chunk_all = ['huge_reinit_chunk', 'run_reinit_chunk']
        calc_free_all = ['run_calc_free']
        get_bitmap_all = ['run_get_bitmap']
        fill_pct_all = ['huge_fill_pct', 'run_fill_pct']

        callers = [
                'memblock_header_none_get_size',
                'block_get_real_size',
                'memblock_from_offset_opt',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, block_size_all)

        callers = [
                'heap_free_chunk_reuse',
                'palloc_heap_action_exec',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, prep_hdr_all)

        callers = [
                'bucket_attach_run',
                'heap_run_into_free_chunk',
                'palloc_reservation_create',
                'palloc_defer_free_create',
                'palloc_defrag',
                'recycler_element_new',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, get_lock_all)

        callers = [
                'container_ravl_insert_block',
                'block_invalidate',
                'alloc_prep_block',
                'palloc_heap_action_on_cancel',
                'palloc_heap_action_on_process',
                'palloc_first',
                'palloc_next',
                'palloc_vg_register_alloc',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, get_user_data_all)

        callers = [
                'bucket_insert_block',
                'memblock_header_legacy_get_size',
                'memblock_header_compact_get_size',
                'memblock_header_legacy_get_extra',
                'memblock_header_compact_get_extra',
                'memblock_header_legacy_get_flags',
                'memblock_header_compact_get_flags',
                'memblock_header_legacy_write',
                'memblock_header_compact_write',
                'memblock_header_legacy_invalidate',
                'memblock_header_compact_invalidate',
                'memblock_header_legacy_reinit',
                'memblock_header_compact_reinit',
                'block_get_user_data',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, get_real_data_all)

        callers = [
                'block_invalidate',
                'alloc_prep_block',
                'palloc_operation',
                'palloc_defrag',
                'palloc_usable_size',
                'palloc_vg_register_alloc',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, get_user_size_all)

        calls = dict_extend(calls, 'alloc_prep_block', write_header_all)

        calls = dict_extend(calls, 'palloc_heap_action_on_cancel', invalidate_all)

        calls = dict_extend(calls, 'heap_get_bestfit_block', ensure_header_type_all)

        calls = dict_extend(calls, 'palloc_vg_register_alloc', reinit_header_all)

        calls = dict_extend(calls, 'heap_vg_open', vg_init_all)

        calls = dict_extend(calls, 'palloc_defrag', get_extra_all)
        calls = dict_extend(calls, 'palloc_extra', get_extra_all)

        calls = dict_extend(calls, 'palloc_defrag', get_flags_all)
        calls = dict_extend(calls, 'palloc_flags', get_flags_all)

        calls = dict_extend(calls, 'bucket_attach_run', iterate_free_all)

        calls = dict_extend(calls, 'heap_zone_foreach_object', iterate_used_all)

        calls = dict_extend(calls, 'heap_reclaim_zone_garbage', reinit_chunk_all)

        calls = dict_extend(calls, 'recycler_element_new', calc_free_all)

        calls = dict_extend(calls, 'palloc_defrag', fill_pct_all)

        # memblock_header_ops
        get_size_all = ['memblock_header_legacy_get_size', 'memblock_header_compact_get_size', 'memblock_header_none_get_size']
        get_extra_all = ['memblock_header_legacy_get_extra', 'memblock_header_compact_get_extra', 'memblock_header_none_get_extra']
        get_flags_all = ['memblock_header_legacy_get_flags', 'memblock_header_compact_get_flags', 'memblock_header_none_get_flags']
        write_all = ['memblock_header_legacy_write', 'memblock_header_compact_write', 'memblock_header_none_write']
        invalidate_all = ['memblock_header_legacy_invalidate', 'memblock_header_compact_invalidate', 'memblock_header_none_invalidate']
        reinit_all = ['memblock_header_legacy_reinit', 'memblock_header_compact_reinit', 'memblock_header_none_reinit']

        callers = [
                'block_get_real_size',
                'memblock_from_offset_opt',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, get_size_all)

        calls = dict_extend(calls, 'block_get_extra', get_extra_all)
        calls = dict_extend(calls, 'block_get_flags', get_flags_all)
        calls = dict_extend(calls, 'block_write_header', write_all)
        calls = dict_extend(calls, 'block_invalidate', invalidate_all)
        calls = dict_extend(calls, 'block_reinit_header', reinit_all)

        # action_funcs
        exec_all = ['palloc_heap_action_exec', 'palloc_mem_action_exec']
        on_cancel_all = ['palloc_heap_action_on_cancel', 'palloc_mem_action_noop']
        on_process_all = ['palloc_heap_action_on_process', 'palloc_mem_action_noop']
        on_unlock_all = ['palloc_heap_action_on_unlock', 'palloc_mem_action_noop']

        calls = dict_extend(calls, 'palloc_exec_actions', exec_all)

        calls = dict_extend(calls, 'palloc_cancel', on_cancel_all)

        calls = dict_extend(calls, 'palloc_exec_actions', on_process_all + on_unlock_all)

        # DAOS used CTLs
        # pmemobj_ctl_get("stats.heap.curr_allocated")
        # pmemobj_ctl_get("stats.heap.run_allocated")
        # pmemobj_ctl_get("stats.heap.run_active")
        get_all = ['ctl__persistent_curr_allocated_read', 'ctl__transient_run_allocated_read', 'ctl__transient_run_active_read']
        calls = dict_extend(calls, 'ctl_exec_query_read', get_all)

        # pmemobj_ctl_set("heap.arenas_assignment_type")
        # pmemobj_ctl_set("heap.alloc_class.new.desc", &pmemslab);
        # pmemobj_ctl_set("stats.enabled", &enabled);
        # pmemobj_ctl_set("stats.enabled", &enabled);
        set_all = ['ctl__arenas_assignment_type_write', 'ctl__desc_write', 'ctl__enabled_write']
        calls = dict_extend(calls, 'ctl_exec_query_write', set_all)

        calls = dict_extend(calls, 'ctl_query', ['ctl_exec_query_read', 'ctl_exec_query_write'])

        # pmem_ops
        persist_all = ['obj_rep_persist', 'obj_norep_persist']
        flush_all = ['obj_rep_flush', 'obj_norep_flush', 'operation_transient_clean', '']
        drain_all = ['obj_rep_drain', 'obj_norep_drain', 'operation_transient_drain']
        memcpy_all = ['obj_rep_memcpy', 'obj_norep_memcpy', 'operation_transient_memcpy']
        memmove_all = ['obj_rep_memmove', 'obj_norep_memmove']
        memset_all = ['obj_rep_memset', 'obj_norep_memset']

        calls = dict_extend(calls, 'pmemops_xpersist', persist_all)
        calls = dict_extend(calls, 'pmemops_xflush', persist_all)
        calls = dict_extend(calls, 'pmemops_drain', drain_all)
        calls = dict_extend(calls, 'pmemops_memcpy', memcpy_all)
        calls = dict_extend(calls, 'pmemops_memmove', memmove_all)
        calls = dict_extend(calls, 'pmemops_memset', memset_all)
        calls = dict_extend(calls, 'ulog_entry_apply', persist_all + flush_all)

        # per-replica functions
        persist_local_all = ['pmem_persist', 'obj_msync_nofail']
        flush_local_all = ['pmem_flush', 'obj_msync_nofail']
        drain_local_all = ['pmem_drain', 'obj_drain_empty']
        memcpy_local_all = ['pmem_memcpy', 'obj_nopmem_memcpy']
        memmove_local_all = ['pmem_memmove', 'obj_nopmem_memmove']
        memset_local_all = ['pmem_memset', 'obj_nopmem_memset']

        calls = dict_extend(calls, 'obj_norep_persist', persist_local_all)
        calls = dict_extend(calls, 'obj_rep_persist', persist_local_all)

        calls = dict_extend(calls, 'obj_norep_flush', flush_local_all)
        calls = dict_extend(calls, 'obj_rep_flush', flush_local_all)

        calls = dict_extend(calls, 'obj_norep_drain', drain_local_all)
        calls = dict_extend(calls, 'obj_rep_drain', drain_local_all)

        callers = [
                'obj_norep_memcpy',
                'obj_rep_memcpy',
                'obj_rep_flush',
                'obj_replicas_check_basic',
        ]
        for caller in callers:
                calls = dict_extend(calls, caller, memcpy_local_all)

        calls = dict_extend(calls, 'obj_norep_memmove', memmove_local_all)
        calls = dict_extend(calls, 'obj_rep_memmove', memmove_local_all)

        calls = dict_extend(calls, 'obj_norep_memset', memset_local_all)
        calls = dict_extend(calls, 'obj_rep_memset', memset_local_all)

        return calls

def get_callees(calls):
        callees = []
        for _, v in calls.items():
                callees.extend(v)
        return list(set(callees))

# XXX
# The way how inlines() function is used shall be changed according to:
# https://github.com/pmem/pmdk/issues/6070
def main():
        extra_calls = core_function_pointers({})
        extra_calls = pmem_function_pointers(extra_calls)
        extra_calls = pmemobj_function_pointers(extra_calls)
        with open("extra_calls.json", "w") as outfile:
                json.dump(extra_calls, outfile, indent = 4)

        extra_calls = inlines(extra_calls)
        # All functions accessed via function pointers have to be provided
        # on top of regular API calls for cflow to process their call stacks.
        extra_entry_points = get_callees(extra_calls)
        extra_entry_points.sort()
        with open("extra_entry_points.txt", "w") as outfile:
                outfile.write("\n".join(extra_entry_points))

if __name__ == '__main__':
        main()



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2024, Intel Corporation

import subprocess # nosec B304
import json
import re

from typing import List, Dict, Any

DEBUG = False
PARSER_DEBUG = False

TOP = '../../src/'
OUTPUT_PATH = TOP + 'test/core_log_max/'
OUTPUT_C = OUTPUT_PATH + 'call_all.c.generated'

NOTICE = """/*
 * This file is automatically generated by utils/call_stack_analysis/log_call_all_generate.py.
 * Please do not modify manually.
 */"""

C_FUNCTION_W_ERRNUM_PREFIX = """
void
call_all_{}(int errnum)
{{
	errno = errnum;
"""

C_FUNCTION_PREFIX = """
void
call_all_{}(void)
{{
"""

C_FUNCTION_SUFFIX = """}
"""

def dump(var: Any, name: str) -> None:
    with open(f'{name}.json', 'w') as outfile:
        json.dump(var, outfile, indent = 4)

def bad_line(reason: str, line: str) -> None:
    print(f'{reason}: {line}')
    exit(1)

# Extract all calls from the code base

def extract_append_code(file_name: str, start_line: int, code: str) -> str:
    if not re.search(f'\.[ch]$', file_name):
        print(f'Unsupported file type: {file_name}')
        exit(1)

    with open(TOP + file_name, 'r') as file:
        # skip to the line
        for _ in range(0, start_line):
            file.readline()
        # look up for the end of the statement
        while True:
            line = file.readline()
            line = line.strip()
            code += " " + line
            if code[-1] == ';':
                break
            if re.search(f'(//|/\*)', code):
                bad_line('Comment found', code)
    return code

IGNORE_FILES = [
    'core/log_internal.h',
    'libpmem2/',
    'libpmempool/',
    'test/',
]

def file_should_be_ignored(file: str) -> bool:
    for ignore in IGNORE_FILES:
        if file.startswith(ignore):
            return True
    return False

def extract_all_calls(func: str) -> List[Dict]:
    # XXX The grep call could be replaced by os.walk() call + for loops over
    # all lines of all files + re.search().
    # In the meantime:
    # B607: Starting a process with a partial executable path - ignored since it
    # is normal way of accessing system utilities.
    # B603: subprocess call - check for execution of untrusted input - there is
    # no way around it. Theoretically, some bad actor could inject faulty grep
    # in the head of the PATH which will lead to us parsing whatever they could
    # benefit from. So, do not commit the produced code automatically.
    returned_output = subprocess.check_output(['grep', '-Irn', func], cwd=TOP) # nosec B607, B603
    string = returned_output.decode("utf-8")
    calls = []
    total = 0
    for line in string.splitlines():
        found = re.search(f'([a-zA-Z0-9_/.-]+):([0-9]+):[ \t]*(.+)', line)
        if found:
            total += 1
            file = found.group(1)
            line_no = found.group(2)
            code = found.group(3)
            # Filter out known odd occurrences.
            if file_should_be_ignored(file):
                continue
            if not code.startswith(func):
                bad_line(f'Does not start with "{func}"', line)
            if code[-1] == '\\': # template's type of ending a line
                if code[-2] == ';': # a single line call can be extracted
                    code = code[:-1]
                else:
                    print(f'SKIP call: Multiline calls in templates are not supported: {line}')
                    continue
            elif code[-1] != ';':
                code = extract_append_code(file, int(line_no), code)
            call = {
                'file': file,
                'line_no': line_no,
                'code': code
            }
            calls.append(call)
        else:
            bad_line('An unexpected line format', line)
    # sort calls by file and line
    def key_func(a: Dict) -> str:
        return a['file'] + a['line_no']
    calls.sort(key=key_func)
    print(f'[{func}] total: {total}, included: {len(calls)}')
    return calls

# Parse a single call's code and tokenize the format string

def parse_token_start(state: Dict, type: str) -> Dict:
    if PARSER_DEBUG:
        print('parse_token_start')
    state['token_start'] = state['pos']
    state['in_' + type] = True
    return state

def parse_token_end(state: Dict) -> Dict:
    if PARSER_DEBUG:
        print('parse_token_end')

    plus = 0
    if state['in_string'] and state['in_literal']:
        bad_line("Both string and literal tokens detected at the same time", state['code'])
    elif state['in_string']:
        _type = 'string'
        plus = 1 # the character ending a string literal belongs to it
    elif state['in_literal']:
        _type = 'literal'
    else:
        bad_line("End of an unknown token detected", state['code'])

    if state['token_start'] is None:
        bad_line("End of a {_type} without the beginning", state['code'])
    state['tokens'].append(state['code'][state['token_start']:(state['pos'] + plus)])
    state['token_start'] = None
    state['in_' + _type] = False
    return state

def parse_call_quote(state: Dict) -> Dict:
    if PARSER_DEBUG:
        print('parse_call_quote')
    if state['in_string']:
        if state['code'][state['pos'] - 1] == '\\': # escaped
            pass
        else:
            state = parse_token_end(state) #  string token ends here
    elif state['in_literal']:
        state = parse_token_end(state) # a literal token ends here
        state = parse_token_start(state, 'string') # a string token starts here...
    else:
        state = parse_token_start(state, 'string') # ... and here as well.
    return state

def parse_call_comma_or_rbracket(state: Dict) -> Dict:
    if PARSER_DEBUG:
        print('parse_call_comma_or_rbracket')
    if state['in_string']:
        pass
    elif state['in_literal']:
        state = parse_token_end(state) # a literal token ends here
        state['end'] = True # the format ends here...
    else:
        state['end'] = True # ... and here
    return state

def parse_call_AZaz09(state: Dict) -> Dict:
    if PARSER_DEBUG:
        print('parse_call_AZaz09')
    if state['in_string']:
        pass
    elif state['in_literal']:
        pass
    else:
        state = parse_token_start(state, 'literal')
    return state

def parse_call_space_or_minus(state: Dict) -> Dict:
    if PARSER_DEBUG:
        print('parse_call_space_or_minus')
    if state['in_string']:
        pass
    elif state['in_literal']:
        state = parse_token_end(state) # a literal token ends here
    return state

def parse_call_odd(state: Dict) -> Dict:
    if PARSER_DEBUG:
        print('parse_call_odd')
    if state['in_string']:
        pass
    else:
        ch = state['code'][state['pos']]
        bad_line('Unexpected "{ch}"', state['code'])
    return state

PARSER_CHAR_ACTION = {
    '"': parse_call_quote,
    ',': parse_call_comma_or_rbracket,
    ')': parse_call_comma_or_rbracket,
    ' ': parse_call_space_or_minus,
    '-': parse_call_space_or_minus,
    '(': parse_call_odd,
    '#': parse_call_odd,
    '%': parse_call_odd,
    ':': parse_call_odd,
    '+': parse_call_odd,
    '\'': parse_call_odd,
    '/': parse_call_odd,
    '.': parse_call_odd,
    '\\': parse_call_odd,
    '<': parse_call_odd,
    '>': parse_call_odd,
    '=': parse_call_odd,
    ';': parse_call_odd,
    '[': parse_call_odd,
    ']': parse_call_odd,
    '!': parse_call_odd,
    '|': parse_call_odd,
    'A-Za-z0-9_': parse_call_AZaz09,
}

def call_get_format_tokens(call):
    code = call['code']
    start = code.find('(')
    state = {
        'code': code,
        'pos': None,
        'end': False,
        'in_string': False,
        'in_literal': False,
        'token_start': None,
        'tokens': []
    }
    for i in range(start + 1, len(code)):
        state['pos'] = i
        ch = code[i]
        if ch in PARSER_CHAR_ACTION.keys():
            state = PARSER_CHAR_ACTION[ch](state)
        elif ch.isalpha() or ch.isdecimal() or ch == '_':
            state = PARSER_CHAR_ACTION['A-Za-z0-9_'](state)
        else:
            bad_line(f'Unsupported character \'{ch}\'', code)
        if state['end']:
            break
    if not state['end']:
        bad_line('Cannot find the end of the format string', code)
    return state['tokens']

# Generate a complete format string

LITERAL_TO_STRING = {
    'PRIx64': 'lx',
    'PRIu64': 'lu'
}

def token_stringify(token: str) -> str:
    if token[0] == '"' and token[-1] == '"':
        return token[1:-1] # strip quotes
    if token in LITERAL_TO_STRING.keys():
        return LITERAL_TO_STRING[token]
    else:
        print(f'Unknown token: "{token}"')
        exit(1)

def format_stringify(tokens: List) -> str:
    tokens_str = [token_stringify(token) for token in tokens]
    return ''.join(tokens_str)

# Identify required arguments

SPECIFIERS = [
    'd',
    'x',
    'u',
    's',
    'p',
    'o',
    'i',
]

FORMAT_SPECIFIER_ARGS = {
    '%s':   '_s',
    '%.8s': '_8s',
    '%x':   '_u',
    '%lx':  '_lu',
    '%#x':  '_u',
    '%d':   '_d',
    '%ld':  '_ld',
    '%u':   '_u',
    '%lu':  '_lu',
    '%zu':  '_zu',
    '%ju':  '_ju',
    '%p':   '_p',
    '%o':   '_u',
    '%i':   '_d',
    '%li':  '_d',
}

def arg_identify(format: str, pos: int):
    format_spec = None
    end_pos = None
    for i in range(pos + 1, len(format)):
        ch = format[i]
        if ch in SPECIFIERS:
            format_spec = format[pos:i + 1]
            end_pos = i + 1
            break
    if format_spec is None:
        visited = format[pos:]
        print(f'Unrecognized specifier: "{visited}"')
        exit(1)
    if format_spec not in FORMAT_SPECIFIER_ARGS.keys():
        print(f'Unrecognized format specifier: "{format_spec}"')
        exit(1)
    return FORMAT_SPECIFIER_ARGS[format_spec], end_pos

def args_parse(format: str) -> List:
    pos = 0
    args = []
    while True:
        pos = format.find('%', pos)
        if pos == -1:
            break
        arg, pos = arg_identify(format, pos)
        args.append(arg)
    return args

# Process raw source code: generate the complete format string and list of args

def calls_process(calls: List) -> List:
    for call in calls:
        call['format_tokens'] = call_get_format_tokens(call)
        call['format_string'] = format_stringify(call['format_tokens'])
        call['args'] = args_parse(call['format_string'])
    return calls

# Generate the call all source file

def init_source_file():
    with open(OUTPUT_C, 'w') as file:
        file.write(NOTICE)

def generate_call(file, func: str, call: Dict) -> str:
    if len(call['args']) > 0:
        args = ', ' + ', '.join(call['args'])
    else:
        args = ''
    file.write(f'\t// src/{call["file"]}\n')
    file.write(f'\t{func}("{call["format_string"]}"{args});\n')

def generate_func_with_errno(func: str, calls: List[Dict]) -> None:
    with open(OUTPUT_C, 'a') as file:
        file.write(C_FUNCTION_W_ERRNUM_PREFIX.format(func))
        for call in calls:
            generate_call(file, func, call)
            file.write('\tUT_ASSERTeq(errno, errnum);\n')
        file.write(C_FUNCTION_SUFFIX)

def generate_func(func: str, calls: List[Dict]) -> None:
    with open(OUTPUT_C, 'a') as file:
        file.write(C_FUNCTION_PREFIX.format(func))
        for call in calls:
            generate_call(file, func, call)
        file.write(C_FUNCTION_SUFFIX)

# Main

API = {
    'void': [
        'CORE_LOG_ERROR_LAST',
        'ERR_WO_ERRNO',
        'CORE_LOG_WARNING',
        'CORE_LOG_ERROR',
        'CORE_LOG_FATAL',
    ],
    'errno': [
        'CORE_LOG_ERROR_W_ERRNO_LAST',
        'ERR_W_ERRNO',
        'CORE_LOG_WARNING_W_ERRNO',
        'CORE_LOG_ERROR_W_ERRNO',
        'CORE_LOG_FATAL_W_ERRNO'
    ]
}

def main():
    global_total = 0
    init_source_file()

    for func in API['void']:
        calls = extract_all_calls(func)
        global_total += len(calls)
        calls = calls_process(calls)
        generate_func(func, calls)

    for func in API['errno']:
        calls = extract_all_calls(func)
        global_total += len(calls)
        calls = calls_process(calls)
        generate_func_with_errno(func, calls)

    print(f'Total: {global_total}')

if __name__ == '__main__':
    main()



#!/usr/bin/python
#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#

import os
from ansible.module_utils.basic import AnsibleModule

DOCUMENTATION = '''
---
module: env
short_description: Module for managing environmental variables on remote hosts
description:
    - "Module for set and unset environmental variables on remote hosts.
       Setting is done by writing export definitions to the file* (this should help with keeping persistence of values
       between reboots) and then execute 'source' command (in order to feed current session with exported variables).
    - ----
    -  * file: the file under the care of this module, specified by 'path' parameter. By default it is a file in
       /etc/profile.d/"
options:
    name:
        description:
            - This is the name of the variable to manage
        required: true
    value:
        description:
            - This is the value of the variable to manage. If 'state'='present', this option is required.
        required: false
    state:
        description:
            - If 'present', ensure that given variable is set to given value. This is the default.
        required: false
        default: present
        choices: [ present ]
    path:
        description:
            - This is the path of a file where export definitions will be written. Usually you don't want to change it.
        required: false
        default: /etc/profile.d/ansible_env_module_definitions.sh
author:
    - dpysx
'''

EXAMPLES = '''
# Set http_proxy:
- name: set http proxy
  env:
    name: http_proxy
    value: http://example.com:80
'''

RETURN = '''
old_value:
    description: Value of variable, before the module was called.
    type: str
'''


# #######################################################################################
# ##########################  General helpers  ##########################################
# #######################################################################################


def check_prerequisites(module, result):
    """Check if passed module and result are valid types from AnsibleModule."""
    if not isinstance(module, AnsibleModule):
        raise TypeError("Given module is not valid!")

    if not isinstance(result, dict):
        raise TypeError("Given result is not valid!")


# #######################################################################################
# ###################  Helpers for dealing with system commands  ########################
# #######################################################################################


def escape_quotes(string):
    """Escape every double quote char with a backslash."""
    return string.replace('"', '\\"')


def run_command(module, result, command):
    """Execute command and return its stdout; when status not zero, fail module execution."""
    check_prerequisites(module, result)

    command = 'bash -lc "' + escape_quotes(command) + '"'

    rc, stdout, stderr = module.run_command(command)

    if rc != 0:
        fail_message = "Command '" + command + "' failed with status code " + str(
            rc) + ", stdout: '" + stdout + "', stderr: '" + stderr + "'."
        module.fail_json(msg=fail_message, **result)

    return stdout


# #######################################################################################
# ######################  Helpers for handling checking env vars  #######################
# #######################################################################################


def get_env_value_or_empty(module, result, name):
    """Obtain value of variable and return it; if not present, return empty string."""
    check_prerequisites(module, result)

    print_variable = 'echo "$' + name + '"'
    variable_value = run_command(module, result, print_variable)

    return variable_value.replace('\n', '')


def is_env_present(module, result, name):
    """Check if variable with given name is present in system."""
    check_prerequisites(module, result)

    env_value = get_env_value_or_empty(module, result, name)
    if env_value == "":
        return False
    else:
        return True


# #######################################################################################
# ##########################  Helpers for file handling  ################################
# #######################################################################################


valid_shebang = "#!/usr/bin/env bash\n"


def read_file_lines(module, result, filename):
    """Read file and return its lines in a form of list."""
    content = []
    try:
        file = open(filename, "r")
        content = file.readlines()
        file.close()
    except IOError:
        fail_message = "Cannot open file '" + filename + "' for reading."
        module.fail_json(msg=fail_message, **result)

    return content


def is_valid_shebang(module, result, filename):
    """Check if file has valid, bash shebang line."""
    check_prerequisites(module, result)

    file_lines = read_file_lines(module, result, filename)

    global valid_shebang
    if len(file_lines) == 0 or file_lines[0] != valid_shebang:
        return False
    else:
        return True


def write_file_lines(module, result, filename, lines):
    """Write all lines to the file."""
    check_prerequisites(module, result)

    changed = False
    if not os.access(filename, os.F_OK):
        changed = True

    try:
        file = open(filename, 'w')
    except IOError:
        fail_message = "Cannot open file '" + filename + "' for write."
        module.fail_json(msg=fail_message, **result)

    try:
        content = []
        for line in lines:
            if not line.endswith('\n'):
                s = line + "\n"
            else:
                s = line
            content.append(s)

        file.writelines(content)
        changed = True
    except:
        fail_message = "Cannot write to file '" + filename + "'."
        module.fail_json(msg=fail_message, **result)
    finally:
        file.close()

    result['changed'] = changed


def write_variable_to_file(module, result, name, value, filename):
    """Write export definition of the variable to the file. Removes possible duplicates."""
    check_prerequisites(module, result)
    global valid_shebang

    content = []
    env_entry_start = "export " + name + "="
    env_entry = env_entry_start + '"' + escape_quotes(value) + '"\n'

    if os.access(filename, os.F_OK):
        content = read_file_lines(module, result, filename)

        if not is_valid_shebang(module, result, filename):
            content.insert(0, valid_shebang)

        # remove all entries for current variable:
        filtered_content = []
        for line in content:
            if not line.startswith(env_entry_start):
                filtered_content.append(line)

        content = filtered_content
        content.append(env_entry)

        write_file_lines(module, result, filename, content)
        result['changed'] = True

    else:
        content.append(valid_shebang)
        content.append(env_entry)
        write_file_lines(module, result, filename, content)
        try:
            os.chmod(filename, 0o755)
            result['changed'] = True
        except:
            fail_message = "Cannot change file '" + filename + "' permissions to 755."
            module.fail_json(msg=fail_message, **result)


# #######################################################################################
# ###############  Helpers for exporting / unset vars to current session  ###############
# #######################################################################################


def import_variables_from_file(module, result, filename):
    """Execute 'source' command with file parameter in order to import definitions to the current session."""
    check_prerequisites(module, result)

    result['changed'] = True

    load_variables_from_file = 'source "' + filename + '"'
    run_command(module, result, load_variables_from_file)


def export_variable(module, result, name, value):
    """Execute 'export' command in order to add exported definition."""
    check_prerequisites(module, result)

    result['changed'] = True
    export_var = "export " + name + '="' + escape_quotes(value) + '"'
    run_command(module, result, export_var)


# #######################################################################################
# ##########################  Main module functions  ####################################
# #######################################################################################


def present_flow(module, result, name, value, filename):
    """Execute module flow for 'present' module mode."""
    check_prerequisites(module, result)

    current_value = get_env_value_or_empty(module, result, name)
    result['old_value'] = current_value

    if value is not None:

        if is_env_present(module, result, name) and current_value == value:
            module.exit_json(**result)

        result['changed'] = True

        if not module.check_mode:
            write_variable_to_file(module, result, name, value, filename)
            import_variables_from_file(module, result, filename)
            export_variable(module, result, name, value)

        module.exit_json(**result)

    else:
        if is_env_present(module, result, name):
            module.exit_json(**result)
        else:
            module.fail_json("variable is not set!", **result)


def run_module():
    """Prepare parameter and return value object, select and execute proper module flow."""

    module_args = dict(
        state=dict(type='str', default='present', choices=['present']),
        path=dict(type='str', default='/etc/profile.d/ansible_env_module_definitions.sh'),
        name=dict(type='str', required=True),
        value=dict(type='str'),
    )

    result = dict(
        changed=False,
        old_value=''
    )

    module = AnsibleModule(
        argument_spec=module_args,
        supports_check_mode=True
    )

    # state = module.params['state']
    name = module.params['name']
    value = module.params['value']
    path = module.params['path']

    # arguments' reasonableness check
    if value is None:
        module.fail_json(msg="There should be 'value' set as well!", **result)

    present_flow(module, result, name, value, path)

    # flow functions should end module's execution; when otherwise, there is some ancient evil in here:
    module.fail_json(msg="Unexpected execution path, this is a bug in this module!", **result)


def main():
    run_module()


if __name__ == '__main__':
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2022, Intel Corporation

from collections import namedtuple

import testframework as t


TC = namedtuple('TC', ['dest', 'src', 'length'])


class PmemMemcpy(t.Test):
    test_type = t.Short
    filesize = 4 * t.MiB
    envs0 = ()
    envs1 = ()
    test_cases = (
        # aligned everything
        TC(dest=0, src=0, length=4096),

        # unaligned dest
        TC(dest=7, src=0, length=4096),

        # unaligned dest, unaligned src
        TC(dest=7, src=9, length=4096),

        # aligned dest, unaligned src
        TC(dest=0, src=9, length=4096)
    )

    def run(self, ctx):
        for env in self.envs0:
            ctx.env[env] = '0'
        for env in self.envs1:
            ctx.env[env] = '1'
        for tc in self.test_cases:
            filepath = ctx.create_holey_file(self.filesize, 'testfile',)
            ctx.exec('pmem_memcpy', filepath, tc.dest, tc.src, tc.length)


class TEST0(PmemMemcpy):
    pass


class TEST1(PmemMemcpy):
    envs1 = ("PMEM_NO_MOVNT",)


class TEST2(PmemMemcpy):
    envs1 = ("PMEM_NO_MOVNT", "PMEM_NO_GENERIC_MEMCPY")


@t.require_architectures('x86_64')
class TEST3(PmemMemcpy):
    envs0 = ("PMEM_AVX512F",)


@t.require_architectures('x86_64')
class TEST4(PmemMemcpy):
    envs0 = ("PMEM_AVX512F", "PMEM_AVX",)


@t.require_architectures('x86_64')
class TEST5(PmemMemcpy):
    envs1 = ("PMEM_MOVDIR64B",)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#

"""
Set of classes that represent the context of single test execution
as well as functions used for setting test requirements and adding
test parameters.

The test context is visible to the RUNTESTS main script
as well as the test user as a Context class.
The context class is constructed with use of 'elements' classes,
that are affiliated with test requirements and test parameters.
"""

import os
import shlex
import sys
import itertools
import shutil
import subprocess as sp
import collections
import contextlib
from inspect import ismethod

import configurator
import futils
from poolset import _Poolset
from python_gdb import GdbProcess
from tools import Tools
from consts import KiB, MiB, HEADER_SIZE
from unsafe_shutdown import UnsafeShutdown
from badblock import BadBlock

try:
    import testconfig
except ImportError:
    sys.exit('Please add valid testconfig.py file - see testconfig.py.example')
config = testconfig.config

try:
    import envconfig
    envconfig = envconfig.config
except ImportError:
    # if file doesn't exist create dummy object
    envconfig = {'GLOBAL_LIB_PATH': '', 'PMEM2_AVX512F_ENABLED': '',
                 'PMEM2_MOVDIR64B_ENABLED': ''}


def expand(*classes):
    """Return flatten list of container classes with removed duplications"""
    return list(set(itertools.chain(*classes)))


class ContextBase:
    """Context basic interface and low-level utilities

    Main context class responsibility is to manage context elements
    appended to it (such as Valgrind, filesystem type, etc.),
    which provide concrete contextualized behavior implementation.

    Attributes:
        _elems (list): list of context elements, such as Valgrind tool,
            filesystem type, etc.
        cmd_prefix (str): test execution command prefix - utilized by
            additional test execution tools, such as Valgrind.
            Defaults to an empty string.
        build (Build): test build

    """

    def __init__(self, build, **elements):
        """
        Args:
            build (Build): test execution build type
            **elements: context elements stored as ContextBase attributes
                named by keyword argument key.
                The element is therefore accessible from ContextBase class.
                Apart from that, below element class methods and attributes
                are handled (if present) by ContextBase class:

                env (dict): environment variables set by the element
                cmd (string): set to cmd_prefix
                setup, check, clean methods: called by corresponding
                    ContextBase methods which are in turn called in
                    Basetest._execute()

        """
        self._elems = []
        self._env = {}
        self.cmd_prefix = ''
        self.build = build

        # for keyword arguments, each kwarg value is set as a
        # context attribute with its key used as an attribute name.
        for k, v in elements.items():
            self.add_ctx_elem(v)
            setattr(self, str(k), v)

    def __getattr__(self, name):
        """
        If context class itself does not have an acquired attribute
        check if one of its elements has it and return it.
        """
        elems_with_attr = [e for e in self._elems if hasattr(e, name)]

        # no context elements with requested attribute found
        if not elems_with_attr:
            raise AttributeError('Neither context nor any of its elements '
                                 'have the attribute "{}"'.format(name))

        # all found elements have the same attribute value
        elif all(e == e for e in elems_with_attr):
            return getattr(elems_with_attr[0], name)

        # more than one element found and they have different attribute values
        else:
            raise AttributeError('Ambiguity while acquiring attribute "{}": '
                                 'implemented by multiple context elements '
                                 'with different values'
                                 .format(name))

    def __str__(self):
        """
        The Context string representation is a concatenation
        (separated by '/') of all its elements string representations.
        """
        s = '{}/'.format(str(self.build))
        for e in self._elems:
            if e:
                s = s + str(e) + '/'
        s = s[:-1]
        return s

    def add_ctx_elem(self, elem):
        """
        Add element to the context. Update context environment variables
        with element's environment variables (if present)
        """
        self._elems.append(elem)

        # if element has a 'name' attribute, it is accessible from Context
        # through it, otherwise its attribute name is generated
        # from its string representation
        if hasattr(elem, 'name'):
            setattr(self, elem.name, elem)
        else:
            setattr(self, str(elem), elem)

        if hasattr(elem, 'env'):
            self.add_env(elem.env)
        if hasattr(elem, 'cmd'):
            if self.cmd_prefix:
                raise ValueError('More than one context element '
                                 'defines command line prefix')
            else:
                self.cmd_prefix = elem.cmd

    def add_env(self, env):
        """Add environment variables to those stored by context"""
        futils.add_env_common(self._env, env)

    def setup(self, *args, **kwargs):
        """
        run setup() method for each context element.
        Ignore error if not implemented
        """
        kwargs['tools'] = self.tools
        for e in self._elems:
            with contextlib.suppress(AttributeError):
                e.setup(*args, **kwargs)

    def check(self, *args, **kwargs):
        """
        run check() method for each context element.
        Ignore error if not implemented
        """
        kwargs['tools'] = self.tools
        for e in self._elems:
            with contextlib.suppress(AttributeError):
                e.check(*args, **kwargs)

    def clean(self, *args, **kwargs):
        """
        run clean() method for each context element.
        Ignore error if not implemented
        """
        kwargs['tools'] = self.tools
        for e in self._elems:
            with contextlib.suppress(AttributeError):
                e.clean(*args, **kwargs)

    @property
    def env(self):
        """Get context environment variables"""
        return self._env

    @property
    def tools(self):
        """Get test tools"""
        return Tools(self.env, self.build)

    @property
    def usc(self):
        return UnsafeShutdown()

    @property
    def badblock(self):
        return BadBlock(self.tools)

    def dump_n_lines(self, file, n=-1):
        """
        Prints last n lines of given log file. Number of lines printed can be
        modified locally by "n" argument or globally by "dump_lines" in
        testconfig.py file. If none of them are provided, default value is 30.
        """
        if n < 0:
            n = config.get('dump_lines', 30)

        file_size = self.get_size(file.name)
        # if file is small enough, read it whole and find last n lines
        if file_size < 100 * MiB:
            lines = list(file)
            length = len(lines)
            if n > length:
                n = length
            lines = lines[-n:]
            lines.insert(0, 'Last {} lines of {} below '
                         '(whole file has {} lines):{}'
                         .format(n, file.name, length, os.linesep))
            for line in lines:
                print(line, end='')
        else:
            # if file is really big, read the last 10KiB and forget about lines
            with open(file.name, 'br') as byte_file:
                byte_file.seek(file_size - 10 * KiB)
                print(byte_file.read().decode('iso_8859_1'))

    def is_devdax(self, path):
        """Checks if given path points to device dax"""
        proc = self.tools.pmemdetect('-d', path)
        if proc.returncode == self.tools.PMEMDETECT_ERROR:
            futils.fail(proc.stdout)
        if proc.returncode == self.tools.PMEMDETECT_TRUE:
            return True
        if proc.returncode == self.tools.PMEMDETECT_FALSE:
            return False
        futils.fail('Unknown value {} returned by pmemdetect'
                    .format(proc.returncode))

    def supports_map_sync(self, path):
        """Checks if MAP_SYNC is supported on a filesystem from given path"""
        proc = self.tools.pmemdetect('-s', path)
        if proc.returncode == self.tools.PMEMDETECT_ERROR:
            futils.fail(proc.stdout)
        if proc.returncode == self.tools.PMEMDETECT_TRUE:
            return True
        if proc.returncode == self.tools.PMEMDETECT_FALSE:
            return False
        futils.fail('Unknown value {} returned by pmemdetect'
                    .format(proc.returncode))

    def get_size(self, path):
        """
        Returns size of the file or dax device.
        Value "2**64 - 1" is checked because pmemdetect
        prints it in case of error.
        """
        proc = self.tools.pmemdetect('-z', path)
        if int(proc.stdout) != 2**64 - 1:
            return int(proc.stdout)
        futils.fail('Could not get size of the file, '
                    'it is inaccessible or does not exist')

    def get_free_space(self, dir="."):
        """Returns free space for current file system"""
        _, _, free = shutil.disk_usage(dir)
        return free


class Context(ContextBase):
    """Manage test execution based on values from context classes"""

    def __init__(self, *args, **kwargs):
        self.conf = configurator.Configurator().config
        self.msg = futils.Message(self.conf.unittest_log_level)
        ContextBase.__init__(self, *args, **kwargs)

    def new_poolset(self, path):
        return _Poolset(path, self)

    def exec(self, cmd, *args, expected_exitcode=0, stderr_file=None,
             stdout_file=None, std_input=None):
        """
        Execute binary in the current test context as a separate process.

        Execution takes place in test cwd and uses environment variables
        stored in Context 'env' attribute. Timeout for the execution
        is set based on the execution configuration.

        Args:
            cmd (str): command to be executed
            *args: Variable length command arguments list
            expected_exitcode (int): if process exit code differs from
                expected, Fail is thrown. Defaults to 0. Ignored
                if set to None.
            stderr_file (str): path to file in which stderr output is
                stored. Stored in a string if None. Defaults to None.
            stdout_file (str): path to file in which stdout output is
                stored. Stored in a string if None. Defaults to None.
            std_input (list): list of strings given to process as an input.
                Defaults to None.

            If neither stderr_file nor stdout_file are set, both outputs
            are merged into single stdout output and stored in a string.
        """
        tmp = self._env.copy()
        futils.add_env_common(tmp, os.environ.copy())

        # change cmd into list for supbrocess type compliance
        cmd = [cmd, ]

        cmd[0] = os.path.join(self.cwd, cmd[0]) + \
            self.build.exesuffix

        if self.valgrind:
            cmd = self.valgrind.cmd + cmd

        # cast all provided args to strings (required by subprocess run())
        # so that exec() can accept args of any printable type
        cmd.extend([str(a) for a in args])

        if self.conf.tracer:
            cmd = shlex.split(self.conf.tracer) + cmd

            # process stdout and stderr are not redirected - this lets running
            # tracer command in an interactive session
            proc = sp.run(cmd, env=tmp, cwd=self.cwd)
        else:
            if stderr_file:
                f = open(os.path.join(self.cwd, stderr_file), 'w')

            # let's create a dictionary of arguments to the run func
            run_kwargs = {
                'env': tmp,
                'cwd': self.cwd,
                'timeout': self.conf.timeout,
                'stdout': sp.PIPE,
                'universal_newlines': True,
                'stderr': sp.STDOUT if stderr_file is None else f}

            if std_input is not None:
                proc_input = ''
                for input in std_input:
                    if not input.endswith('\n'):
                        input = ''.join([input, '\n'])
                    proc_input = ''.join([proc_input, input])
                proc = sp.run(cmd, **run_kwargs, input=proc_input)
            else:
                proc = sp.run(cmd, **run_kwargs)

            if stderr_file:
                f.close()

        if expected_exitcode is not None and \
           proc.returncode != expected_exitcode:
            futils.fail(proc.stdout, exit_code=proc.returncode)

        if stdout_file is not None:
            with open(os.path.join(self.cwd, stdout_file), 'w') as f:
                f.write(proc.stdout)

        self.msg.print_verbose(proc.stdout)

    def create_holey_file(self, size, path, mode=None):
        """
        Create a new file in test testdir
        with the selected size and name.
        """
        filepath = os.path.join(self.testdir, path)
        with open(filepath, 'w') as f:
            if size > 0:
                f.seek(size - 1)
                f.write('\0')
        if mode is not None:
            os.chmod(filepath, mode)
        return filepath

    def create_non_zero_file(self, size, path, mode=None):
        """
        Create a new non-zeroed file in test testdir
        with the selected size and name.
        """
        filepath = os.path.join(self.testdir, path)
        with open(filepath, 'w') as f:
            f.write('\132' * size)
        if mode is not None:
            os.chmod(filepath, mode)
        return filepath

    def create_zeroed_hdr_file(self, size, path, mode=None):
        """
        Create a new non-zeroed file with a zeroed header and the selected
        size and name.
        """
        filepath = os.path.join(self.testdir, path)
        with open(filepath, 'w') as f:
            f.write('\0' * HEADER_SIZE)
            f.write('\132' * (size - HEADER_SIZE))
        if mode is not None:
            os.chmod(filepath, mode)
        return filepath

    def require_free_space(self, space):
        """
        Skip test if not enough free space is available in test testdir.
        """
        if self.get_free_space(self.testdir) < space:
            futils.skip('Not enough free space ({} MiB required)'
                        .format(space / MiB))

    def mkdirs(self, path, mode=None):
        """
        Creates directory in test testdir along with all parent
        directories required. In the case given path already exists
        do nothing.
        """
        dirpath = os.path.join(self.testdir, path)
        if mode is None:
            os.makedirs(dirpath, exist_ok=True)
        else:
            os.makedirs(dirpath, mode, exist_ok=True)

    def gdb(self, testnum, path_to_exec, args_to_exec='',
            gdb_options='', timeout=15):
        """Get gdb process"""
        return GdbProcess(self.env, self.cwd, testnum, path_to_exec,
                          args_to_exec, gdb_options, timeout)


class CtxType(type):
    """
    Metaclass for context classes that can stand for multiple classes

    Attributes:
        is_preferred (bool): if True, context class is preferred
            to be chosen in case of Any execution setting.
        explicit (bool): if True, the context class is executed
            only if explicitly requested by test implementation,
            i. e. it is ignored if specified only in external
            configuration.
        includes: list of context classes included by this class,
            e. g. Short and Medium test types are included by
            Check test type

    """
    def __init__(cls, name, bases, dct):
        type.__init__(cls, name, bases, dct)

        # it is supposed that context class has a user defined base class
        if cls.__base__.__name__ == 'object':
            return

        # context class preferred to be used when 'Any' test option is provided
        if not hasattr(cls, 'preferred'):
            cls.is_preferred = False

        # if explicit, run only if the test case explicitly specifies this
        # context in its own configuration
        if not hasattr(cls, 'explicit'):
            cls.explicit = False

        if hasattr(cls, 'includes'):
            cls.includes = expand(*cls.includes)
        else:
            # class that does not include other classes, includes itself
            cls.includes = [cls]

        # set instance (not class) method
        setattr(cls, '__repr__', lambda cls: cls.__class__.__name__.lower())

    def __repr__(cls):
        return cls.__name__.lower()

    def __str__(cls):
        return cls.__name__.lower()

    def __iter__(cls):
        for c in cls.includes:
            yield c

    def factory(cls, conf, *classes):
        return [c(conf) for c in expand(*classes)]


def filter_contexts(config_ctx, test_ctx):
    """
    Return contexts that should be used in execution based on
    contexts provided by config and test case.
    """
    if not test_ctx:
        return [c for c in config_ctx if not c.explicit]
    return [c for c in config_ctx if c in test_ctx]


def str_to_ctx_common(val, ctx_base_type):
    def class_from_string(name, base):
        if name == 'all':
            return base.__subclasses__()

        try:
            return next(b for b in base.__subclasses__()
                        if str(b) == name.lower())
        except StopIteration:
            print('Invalid context value: "{}".'.format(name))
            raise

    if isinstance(val, list):
        classes = [class_from_string(cl, ctx_base_type) for cl in val]
        return expand(*classes)
    else:
        return expand(class_from_string(val, ctx_base_type))


def _req_prefix(attr):
    """
    Add prefix to requirement attribute name. Used to mitigate the risk
    of name clashes.
    """
    return '_require{}'.format(attr)


def add_requirement(tc, attr, value, **kwargs):
    """
    Add requirement to the test

    Requirements are added to the test as attributes dynamically
    appended to the test class. Their names are
    additionally mangled with _req_prefix() function
    to avoid accidental overwriting of any existing
    test class attributes as well as unnecessary access by the
    test user - these attributes are meant to be used directly
    only by requirements handling functions.

    Args:
        tc (BaseTest): test case to which the requirement
            is added
        attr (str): attribute name of the requirement.
            additionally mangled with _req_prefix() function.
        value: requirement value
        **kwargs: additional keyword arguments that may be used
            by requirement implementation
    """
    setattr(tc, _req_prefix(attr), value)
    setattr(tc, _req_prefix('{}_kwargs'.format(attr)), kwargs)


def get_requirement(tc, attr, default):
    """
    Get test requirement added with add_requirement() function
    with specific attribute name.

    Args:
        tc (BaseTest): test case class into which requirements were
            added
        attr (str): requirement attribute name - as previously added by
            add_requirement() function
        default: default requirement value if not found

    Returns:
        requirement value, requirement keyword arguments
            (as provided by 'kwargs' argument to add_requirement() function)
    """
    ret_val = default
    ret_kwargs = {}
    try:
        ret_val = getattr(tc, _req_prefix(attr))
        ret_kwargs = getattr(tc, _req_prefix('{}_kwargs'.format(attr)))
    except AttributeError:
        pass

    return ret_val, ret_kwargs


class TestParam:
    """
    TestParam manages single test parameter provided to test with add_params()
    decorator.
    Test parameters are added to context class as its elements - therefore the
    'value' object may implement interface handled by ContextBase class, such
    as setup() and clean()
    """
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __str__(self):
        return '{}: {}'.format(self.name, self.value)

    def __call__(self):
        return self.value

    def setup(self, **kwargs):
        if hasattr(self.value, 'setup') and ismethod(self.value.setup):
            self.value.setup(**kwargs)

    def check(self, **kwargs):
        if hasattr(self.value, 'check') and ismethod(self.value.check):
            self.value.check(**kwargs)

    def clean(self, **kwargs):
        if hasattr(self.value, 'clean') and ismethod(self.value.clean):
            self.value.clean(**kwargs)


PARAMS_ATTR = '_params_'


def add_params(name, values):
    """
    Add parameters to the test case.

    Parameters will be accessible from test through context 'name' attribute.
    The value of single parameter can be accessed by calling this attribute,
    e.g. for:
    add_params('param', params_list)
    single parameter provided within 'params_list' is accessible
    by calling ctx.param().
    """
    params = [TestParam(name, v) for v in values]

    def wrapped(tc):
        if hasattr(tc, PARAMS_ATTR):
            getattr(tc, PARAMS_ATTR).update({name: params})
        else:
            setattr(tc, PARAMS_ATTR, {name: params})
        return tc

    return wrapped


def get_params(tc):
    """Get parameters added to test case"""
    try:
        return getattr(tc, PARAMS_ATTR)
    except AttributeError:
        return {}


class Any:
    """
    The test context attribute signifying that specific context value is not
    relevant for the test outcome and it should be run only once in some
    viable context
    """
    @classmethod
    def get(cls, conf_ctx):
        """Get specific context value to be run"""
        for c in conf_ctx:
            if c.is_preferred:
                # pick preferred if found
                return c
        # if no preferred is found, pick the first non-explicit one
        ret = [c for c in conf_ctx if not c.explicit]
        if ret:
            return ret[0]
        else:
            config = configurator.Configurator().config
            msg = futils.Message(config.unittest_log_level)
            msg.print_verbose('No valid "Any" context found')
            return None


class _NoContext(collections.UserList):
    """
    May be safely returned by context element class 'filter' method if
    its items are not required during test execution
    (e. g. no dax devices required by test)
    """

    def __init__(self):
        self.data = []
        self.data.append(False)

    def __bool__(self):
        return False


NO_CONTEXT = _NoContext()



#!/usr/bin/env python3

# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2024, Intel Corporation

import argparse
import json
import pandas as pd
import subprocess


PARSER = argparse.ArgumentParser()
PARSER.add_argument('--reference', metavar='LD_LIBRARY_PATH', required=True,
        help='LD_LIBRARY_PATH where the first version of PMDK is built')
PARSER.add_argument('--rival', metavar='LD_LIBRARY_PATH', required=True,
        help='LD_LIBRARY_PATH where the second version of PMDK is built')
PARSER.add_argument('-c', '--config', required=True,
        help='Name of the .cfg file to use')
PARSER.add_argument('-s', '--scenario', required=True,
        help='Name of the scenario to run')
PARSER.add_argument('-p', '--pmem_path', required=True,
        help='PMEM-mounted directory to use')


COLUMNS_COMBINE = [
        'ops-per-second[1/sec]',
        'total-avg[sec]',
        'total-max[sec]',
        'total-min[sec]',
        'total-median[sec]',
        'total-std-dev[sec]',
        'latency-avg[nsec]',
        'latency-min[nsec]',
        'latency-max[nsec]',
        'latency-std-dev[nsec]',
        'latency-pctl-50.0%[nsec]',
        'latency-pctl-99.0%[nsec]',
        'latency-pctl-99.9%[nsec]',
]


COLUMNS_COPY = [
        'threads',
        'ops-per-thread',
        'data-size',
        'seed',
        'repeats',
        'thread-affinity',
        'main-affinity',
        'min-exe-time',
        'random',
        'min-size',
        'type-number',
        'operation',
        'lib',
        'nestings',
        'type',
        'max-key',
        'external-tx',
        'alloc',
]


def get_numa_node(pmem_path: str) -> int:
        """Find the NUMA node of the device mounted at the specified path"""
        # no slash at the end of the path
        if pmem_path[-1] == '/':
                pmem_path = pmem_path[0:-1]
        # find the line describing the mount point
        mount = subprocess.getoutput('mount')
        mount_line = None
        for line in mount.splitlines():
                if pmem_path in line:
                        mount_line = line
                        break
        if mount_line is None:
                print(mount)
                raise Exception(f'Can not find the mounted PMEM device for: {pmem_path}')
        # Extract blockdev from the found line e.g.
        # /dev/pmem1 on /mnt/pmem1 type ext4 (rw,noatime,seclabel,nodelalloc,dax=always)
        # - Linux device is the first part of the line e.g. /dev/pmem1
        # - blockdev is the name of the device file e.g. pmem1
        blockdev = mount_line.split(' ')[0].split('/')[2]
        # Find the NDCTL namespace of the blockdev and extract its NUMA node
        numa_node = None
        namespaces = json.loads(subprocess.getoutput('ndctl list -v'))
        for namespace in namespaces:
                if namespace['mode'] == 'fsdax' and namespace['blockdev'] == blockdev:
                        numa_node = namespace['numa_node']
                        break
        if numa_node is None:
                print(namespaces)
                raise Exception(f'Can not find the namespace: mode == "fsdax" and blockdev == "{blockdev}"')
        return numa_node


IDX_TO_NAME = ['ref', 'riv']


def output_name(args: argparse.Namespace, idx: int) -> str:
        """Generate a file name for an output ref or riv"""
        return f'{args.config}__{args.scenario}_{IDX_TO_NAME[idx]}.csv'


def run(numa_node: int, args: argparse.Namespace, idx: int, ld_library_path: str) -> None:
        """Run PMEMBENCH according to the provided parameters"""
        config = f'src/benchmarks/{args.config}.cfg'
        file = f'{args.pmem_path}/testfile.obj'
        cmd = f'numactl --cpunodebind {numa_node} --localalloc ./src/benchmarks/pmembench {config} {args.scenario} --file {file}'
        env = {'LD_LIBRARY_PATH': ld_library_path}
        result = subprocess.run(cmd, env=env, stdout=subprocess.PIPE,
                stderr=subprocess.PIPE, encoding='utf-8', shell=True)
        # generate the file with the output
        if result.returncode == 0:
                # drop the first line e.g. obj_rbtree_map_insert: map_insert [1]
                out = ''.join(result.stdout.splitlines(keepends=True)[1:])
        else:
                out = result.stdout
        with open(output_name(args, idx), 'w') as output:
                output.write(out)
        # validate the run
        if result.returncode != 0:
                print(result.stdout)
                print(result.stderr)
                exit(result.returncode)


def column_name(column: str, idx: int) -> str:
        """Generate a column name with ref or rev infix"""
        return f'-{IDX_TO_NAME[idx]}['.join(column.split('['))


def combine(args: argparse.Namespace) -> None:
        """"
        Combine outputs from the reference and rival runs.

        Output data files:
        - combined - contains data from ref and riv together with a normalized
          difference between them.
        - diff - just a normalized difference between ref and riv
        """
        dfs = [pd.read_csv(output_name(args, idx), sep=';') for idx in range(2)]
        combined = pd.DataFrame()
        diff = pd.DataFrame()
        for column in COLUMNS_COMBINE:
                # Copy columns to combine from both ref and riv
                for idx in range(2):
                        combined[column_name(column, idx)] = dfs[idx][column]
                diff_column = f'{column}-diff'
                # Normalized difference between ref and riv:
                # diff = (riv - ref) / ref
                # Both output data frames contains diff columns.
                combined[diff_column] = (dfs[1][column] / dfs[0][column] - 1)
                diff[diff_column] = combined[diff_column]
        for column in COLUMNS_COPY:
                if column in dfs[0].columns:
                        # These columns are identical in both data frames
                        # so they can be copied from either data frame ref or riv.
                        combined[column] = dfs[0][column]
                        diff[column] = dfs[0][column]
        # Write the generated data frames to CSV files.
        prefix = f'{args.config}__{args.scenario}'
        combined.to_csv(f'{prefix}_combined.csv', index=False, float_format='%.3f')
        diff.to_csv(f'{prefix}_diff.csv', index=False, float_format='%.3f')


def main():
        args = PARSER.parse_args()
        numa_node = get_numa_node(args.pmem_path)
        for idx, ld_library_path in enumerate([args.reference, args.rival]):
                run(numa_node, args, idx, ld_library_path)
        combine(args)


if __name__ == '__main__':
        main()



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2022, Intel Corporation


import testframework as t
from testframework import granularity as g


class BASIC(t.Test):
    test_type = t.Medium

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile1')
        ctx.exec('obj_heap_reopen', filepath)


@g.require_granularity(g.BYTE, g.CACHELINE)
class TEST0(BASIC):
    pass



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#


import testframework as t


class TEST0(t.Test):
    test_type = t.Short

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile')
        ctx.exec('pmem2_source_alignment',
                 'test_get_alignment_success', filepath)


class PMEM2_SOURCE_ALIGNMENT_DEV_DAX(t.Test):
    test_type = t.Short
    test_case = "test_get_alignment_success"

    def run(self, ctx):
        dd = ctx.devdaxes.devdax
        ctx.exec('pmem2_source_alignment',
                 self.test_case, dd.path, dd.alignment)


@t.require_devdax(t.DevDax('devdax', alignment=2 * t.MiB))
class TEST1(PMEM2_SOURCE_ALIGNMENT_DEV_DAX):
    pass


@t.require_devdax(t.DevDax('devdax', alignment=4 * t.KiB))
class TEST2(PMEM2_SOURCE_ALIGNMENT_DEV_DAX):
    pass



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation
#


import testframework as t
from testframework import granularity as g


@g.require_granularity(g.ANY)
class Pmem2_from_existing(t.Test):
    test_type = t.Short

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile1')
        ctx.exec('pmem2_map_from_existing', self.test_case, filepath)


class TEST0(Pmem2_from_existing):
    """try to create two the same mappings"""
    test_case = "test_two_same_mappings"


class TEST1(Pmem2_from_existing):
    """try to map which overlap bottom part of existing mapping"""
    test_case = "test_mapping_overlap_bottom"


class TEST2(Pmem2_from_existing):
    """try to map which overlap upper part of existing mapping"""
    test_case = "test_mapping_overlap_upper"


class TEST3(Pmem2_from_existing):
    """inject enomem in to allocation of map object"""
    labels = ['fault_injection']
    test_case = "test_map_allocation_enomem"


class TEST4(Pmem2_from_existing):
    """inject enomem during adding map to ravl"""
    test_case = "test_register_mapping_enomem"



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#
"""External tools integration"""

import os
import json
import subprocess as sp

import futils

try:
    import envconfig
    envconfig = envconfig.config
except ImportError:
    # if file doesn't exist create dummy object
    envconfig = {'GLOBAL_LIB_PATH': '', 'PMEM2_AVX512F_ENABLED': '',
                 'PMEM2_MOVDIR64B_ENABLED': ''}


class Tools:
    PMEMDETECT_FALSE = 0
    PMEMDETECT_TRUE = 1
    PMEMDETECT_ERROR = 2

    def __init__(self, env, build):
        self.env = env
        self.build = build
        global_lib_path = envconfig['GLOBAL_LIB_PATH']

        futils.add_env_common(self.env,
                              {'LD_LIBRARY_PATH': global_lib_path})
        if build is not None:
            futils.add_env_common(self.env,
                                  {'LD_LIBRARY_PATH': build.libdir})

    def _run_test_tool(self, name, *args):
        exe = futils.get_test_tool_path(self.build, name)

        return sp.run([exe, *args], env=self.env, stdout=sp.PIPE,
                      stderr=sp.STDOUT, universal_newlines=True)

    def pmemdetect(self, *args):
        return self._run_test_tool('pmemdetect', *args)

    def gran_detecto(self, *args):
        return self._run_test_tool('gran_detecto', *args)

    def extents(self, *args):
        return self._run_test_tool('extents', *args)

    def cpufd(self):
        return self._run_test_tool('cpufd')

    def mapexec(self, *args):
        return self._run_test_tool('mapexec', *args)

    def usc_permission_check(self, *args):
        return self._run_test_tool('usc_permission_check', *args)


class Ndctl:
    """ndctl CLI handle

    Attributes:
        version (str): ndctl version
        ndctl_list_output (dict): output of 'ndctl list' command
            decoded from JSON into dictionary
    """
    def __init__(self):

        self.version = self._get_ndctl_version()
        self.ndctl_list_output = self._cmd_out_to_json('list', '-RBNDMv')

    def _cmd_out_to_json(self, *args):
        cmd = ['ndctl', *args]
        cmd_as_str = ' '.join(cmd)
        proc = sp.run(cmd, stdout=sp.PIPE, stderr=sp.STDOUT,
                      universal_newlines=True)
        if proc.returncode != 0:
            raise futils.Fail('"{}" failed:{}{}'.format(cmd_as_str, os.linesep,
                                                        proc.stdout))
        try:
            out = json.loads(proc.stdout)
        except json.JSONDecodeError:
            raise futils.Fail('invalid "{}" output (could '
                              'not read as JSON): {}'.format(cmd_as_str,
                                                             proc.stdout))
        else:
            return out

    def _get_ndctl_version(self):
        """
        Get ndctl version.

        Acquiring the version is simultaneously used as a check whether
        the ndctl is installed on the system.

        Returns:
            ndctl version (str)

        """
        proc = sp.run(['ndctl', '--version'], stdout=sp.PIPE, stderr=sp.STDOUT)
        if proc.returncode != 0:
            raise futils.Fail('checking if ndctl exists failed:{}{}'
                              .format(os.linesep, proc.stdout))

        version = proc.stdout.strip()
        return version

    def _get_dev_region(self, dev_path):
        devtypes = ('blockdev', 'chardev')

        for r in self.ndctl_list_output['regions']:
            for d in r['namespaces']:
                for dt in devtypes:
                    if dt in d and os.path.join('/dev', d[dt]) == dev_path:
                        return r

    def _get_dev_info(self, dev_path):
        """
        Get ndctl information about the given device.
        Returns dictionary associated with device.
        """
        dev = None

        # Possible viable device types as shown with
        # 'ndctl list' output
        devtypes = ('blockdev', 'chardev')

        list = self._cmd_out_to_json('list')
        for d in list:
            for dt in devtypes:
                if dt in d and os.path.join('/dev', d[dt]) == dev_path:
                    dev = d

        if not dev:
            raise futils.Fail('ndctl does not recognize the device: "{}"'
                              .format(dev_path))
        return dev

    # for ndctl v63 we need to parse ndctl list in a different way than for v64
    def _get_dev_info_63(self, dev_path):
        dev = None
        devtype = 'chardev'
        daxreg = 'daxregion'

        list = self._cmd_out_to_json('list')
        for d in list:
            if daxreg in d:
                devices = d[daxreg]['devices']
                for device in devices:
                    if devtype in device and \
                            os.path.join('/dev', device[devtype]) == dev_path:
                        # only params from daxreg are interested at this point,
                        # other values are read by _get_dev_info() earlier
                        dev = d[daxreg]

        if not dev:
            raise futils.Fail('ndctl does not recognize the device: "{}"'
                              .format(dev_path))

        return dev

    def _get_dev_param(self, dev_path, param):
        """
        Acquire device parameter from 'ndctl list' output.

        Args:
            dev_path (str): path of device
            param (str): parameter

        """
        p = None
        dev = self._get_dev_info(dev_path)

        try:
            p = dev[param]
        except KeyError:
            dev = self._get_dev_info_63(dev_path)
            p = dev[param]

        return p

    def _get_emulated_devices(self):
        # provider of emulated pmem
        emulated_pmem_provider = "e820"

        # important keys from ndctl output list
        blkdev = 'blockdev'
        chrdev = 'chardev'
        daxreg = 'daxregion'
        devs = 'devices'

        emulated_devices = []
        for bus in self.ndctl_list_output:
            if bus['provider'] == emulated_pmem_provider:
                for reg in bus['regions']:
                    for ns in reg['namespaces']:
                        if blkdev in ns:
                            edev = ns[blkdev]
                        elif daxreg in ns:
                            for dev in ns[daxreg][devs]:
                                edev = dev[chrdev]
                        else:
                            edev = None

                        emulated_devices.append(edev)

        return emulated_devices

    def _get_path_device(self, path):
        blockdev = futils.run_command("df -P {} | tail -n 1 | cut -f 1 -d ' '"
                                      .format(path)).strip().decode('UTF8')
        if blockdev == "devtmpfs":
            device = path  # devdax
        else:
            device = blockdev  # fsdax

        return device

    def get_dev_size(self, dev_path):
        return int(self._get_dev_param(dev_path, 'size'))

    def get_dev_alignment(self, dev_path):
        return int(self._get_dev_param(dev_path, 'align'))

    def get_dev_mode(self, dev_path):
        return self._get_dev_param(dev_path, 'mode')

    def get_dev_sector_size(self, dev_path):
        return self._get_dev_param(dev_path, 'sector_size')

    def get_dev_namespace(self, dev_path):
        return self._get_dev_param(dev_path, 'dev')

    def get_dev_bb_count(self, dev_path):
        result = self._get_dev_param(dev_path, 'badblock_count')
        return result

    def is_devdax(self, dev_path):
        return self.get_dev_mode(dev_path) == 'devdax'

    def is_fsdax(self, dev_path):
        return self.get_dev_mode(dev_path) == 'fsdax'

    def is_emulated(self, dev_path):
        emulated_devices = self._get_emulated_devices()
        dev = self._get_path_device(dev_path).replace('/dev/', '')

        return (dev in emulated_devices)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2018-2022, Intel Corporation

import subprocess  # nosec
from ctypes import cdll, c_char_p, c_int
from loggingfacility import LoggingBase
from os import path
from sys import exit

checkers = ["prog", "lib"]


class ConsistencyCheckerBase:
    """
    Base class for consistency checker classes.
    Checker of each type should implement check_consistency method.
    """

    def check_consistency(self, filename):
        pass


class LibChecker(ConsistencyCheckerBase):
    """
    Allows registration of a consistency checking function and verifying
    the consistency of a file.

    The function has to be in a shared library. It is then used to check
    consistency of an arbitrary file. The function has to take a file name
    as the only parameter and return an int: 0 for inconsistent, 1 for
    consistent. The prototype of the function:

        int func_name(const char* file_name)
    """

    def __init__(self, library_name, func_name, logger=None):
        """
        Loads the consistency checking function from the given library.

        :param library_name: The full name of the library.
        :type library_name: str
        :param func_name: The name of the consistency
                          checking function within the library.
        :type func_name: str
        :param logger: logger handle, default: empty logger (LoggingBase)
        :type logger: subclass of :class:`LoggingBase`
        :return: None
        """
        self._lib_name = library_name
        self._lib_func_name = func_name
        self._lib_func = getattr(cdll.LoadLibrary(library_name), func_name)
        self._lib_func.argtypes = [c_char_p]
        self._lib_func.restype = c_int
        self._logger = logger or LoggingBase()

    def check_consistency(self, filename):
        """
        Checks the consistency of a given file
        using the previously loaded function.

        :param filename: The full name of the file to be checked.
        :type filename: str
        :return: 1 if file is consistent, 0 otherwise.
        :rtype: int
        :raises: Generic exception, when no function has been loaded.
        """
        if self._lib_func is None:
            raise RuntimeError(
                "Consistency check function {} not loaded".format(
                    self._lib_func_name
                )
            )

        self._logger.debug(
            "Consistency check function: {0}.{1}({2})".format(
                self._lib_name, self._lib_func_name, filename
            )
        )
        return self._lib_func(filename)


class ProgChecker(ConsistencyCheckerBase):
    """
    Allows registration of a consistency checking program and verifying
    the consistency of a file.

    The binary executed with its argument is used to check consistency
    of an arbitrary file. The program has to take a file name as the
    last parameter and return an int: 0 for inconsistent, 1 for consistent.
    """

    def __init__(self, bin_path, bin_args, logger=None):
        """
        Loads the consistency checking binary and its arguments required
        to verify the consistency of a file.

        :param bin_path: The full path of the binary.
        :type bin_path: str
        :param bin_args: Binary's arguments to run consistency check.
        :type bin_args: str
        :param logger: logger handle, default: empty logger (LoggingBase)
        :type logger: subclass of :class:`LoggingBase`
        :return: None
        """
        self._bin_path = bin_path
        self._bin_cmd = bin_args
        self._logger = logger or LoggingBase()

    def check_consistency(self, filename):
        """
        Checks the consistency of a given file
        using the previously loaded function.

        :param filename: The full name of the file to be checked.
        :type filename: str
        :return: 1 if file is consistent, 0 otherwise.
        :rtype: int
        :raises: Generic exception, when no function has been loaded.
        """
        if self._bin_path is None or self._bin_cmd is None:
            raise RuntimeError("consistency check handle not set")

        cmd = "{0} {1} {2}".format(self._bin_path, self._bin_cmd, filename)
        self._logger.debug("Consistency check program command: {}".format(cmd))
        """
        We mark the call of this command as 'nosec' (for Bandit scan) because
        pmreorder entirely relies on the execution of checkers, which are
        user-developed programs. Therefore, it is the user's responsibility
        to provide safe input as a consistency checker.
        """
        return subprocess.call(cmd, shell=True)  # nosec


def get_checker(checker_type, checker_path_args, func_name, logger=None):
    """
    Returns checker instance, based on the checker type.

    :param checker_type: Type of the checker, supported types: "prog", "lib".
    :type checker_type: str
    :param checker_path_args: Checker's arguments for consistency check.
    :type checker_path_args: str
    :param func_name: Name of the checker function, given only if "lib" type.
    :type func_name: str
    :param logger: logger handle, default: None
    :type logger: subclass of :class:`LoggingBase`
    :return: subclass of :class:`ConsistencyCheckerBase`
    """
    checker_path_args = checker_path_args.split(" ", 1)
    checker_path = checker_path_args[0]

    # check for params
    if len(checker_path_args) > 1:
        args = checker_path_args[1]
    else:
        args = ""

    if not path.exists(checker_path):
        print("Invalid path: " + checker_path)
        exit(1)

    checker = None
    if checker_type == "prog":
        checker = ProgChecker(checker_path, args, logger)
    elif checker_type == "lib":
        checker = LibChecker(checker_path, func_name, logger)

    return checker



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation
#
"""Various requirements"""

import os
from shutil import which
import subprocess as sp

import configurator as conf
import context as ctx
import futils


NDCTL_MIN_VERSION = '63'


class Requirements:
    """Various requirements"""
    def __init__(self):
        self.cfg = conf.Configurator().config

    def _check_pkgconfig(self, pkg, min_ver):
        """
        Run pkg-config to check if a package is installed
        """
        proc = sp.run(['pkg-config', pkg, '--atleast-version', min_ver],
                      stdout=sp.PIPE, stderr=sp.STDOUT)
        return proc.returncode == 0

    def _is_ndctl_enabled(self):
        path = futils.get_tool_path(ctx, "pmempool")
        s = sp.check_output(["strings", path])
        if "compiled with libndctl" in str(s):
            return True

        return False

    def _check_is_admin(self):
        cmd = "sudo -n id -u"
        fail_msg = "checking user id failed"

        cmd_output = futils.run_command(cmd, fail_msg)
        cmd_output = cmd_output.strip().decode('UTF8')
        if cmd_output != '0':
            return False

        return True

    def check_ndctl(self):
        is_ndctl = self._check_pkgconfig('libndctl', NDCTL_MIN_VERSION)
        if not is_ndctl:
            raise futils.Skip('libndctl (>=v{}) is not installed'
                              .format(NDCTL_MIN_VERSION))

    def check_ndctl_enable(self):
        if self._is_ndctl_enabled() is False:
            raise futils.Skip('ndctl is disabled - binary not '
                              'compiled with libndctl')

    def check_namespace(self):
        cmd = ['ndctl', 'list']
        cmd_as_str = ' '.join(cmd)
        try:
            proc = sp.run(cmd, stdout=sp.PIPE, stderr=sp.STDOUT,
                          universal_newlines=True)
        except (OSError):
            raise futils.Fail('ndctl is not installed')
        if proc.returncode != 0:
            raise futils.Fail('"{}" failed:{}{}'.format(cmd_as_str, os.linesep,
                                                        proc.stdout))
        if not proc.stdout or proc.stdout.isspace():
            raise futils.Skip('no ndctl namespace set')

    def _check_ndctl_req_is_met(self, tc):
        """
        Check if all conditions for the ndctl requirement are met
        """
        require_ndctl, kwargs = ctx.get_requirement(tc, 'require_ndctl', ())
        if not require_ndctl:
            return True

        self.check_ndctl_enable()
        self.check_ndctl()

        if kwargs.get('require_namespace', False):
            self.check_namespace()

        return True

    def _check_admin_req_is_met(self, tc):
        """
        Check if all conditions for the admin requirement are met
        """
        require_admin, _ = ctx.get_requirement(tc, 'require_admin', ())
        if not require_admin:
            # admin is not required
            return True

        if not self.cfg.enable_admin_tests:
            raise futils.Skip('admin tests are not enabled in config '
                              '(enable_admin_tests)')

        if not self._check_is_admin():
            raise futils.Fail('Error: admin tests are enabled in config, '
                              'but the user does not have administrative '
                              'privileges')
        # user is admin
        return True

    def check_if_all_requirements_are_met(self, tc):
        if not self._check_ndctl_req_is_met(tc):
            return False
        if not self._check_admin_req_is_met(tc):
            return False

        """More requirements can be checked here"""
        return True


def require_ndctl(**kwargs):
    """
    Add requirement to run test only if ndctl is installed.
    Optionally a namespace requirement can be enabled.

    Used as a test class decorator.

    Args:
        kwargs: optional keyword arguments
    """
    valid_kwarg_keys = ['require_namespace']

    # check if all provided keys are valid
    for key in kwargs.keys():
        if key not in valid_kwarg_keys:
            raise KeyError('provided key {} is invalid'.format(key))

    # namespace requirement
    namespace_kw = 'require_namespace'
    namespace_val = kwargs.get(namespace_kw, False)
    if not isinstance(namespace_val, bool):
        raise ValueError('provided value {} for optional key {} is invalid'
                         .format(namespace_val, namespace_kw))

    def wrapped(tc):
        ctx.add_requirement(tc, 'require_ndctl', True, **kwargs)
        return tc
    return wrapped


def require_admin(tc):
    """
    Disable test if "enable_admin_tests" configuration is not set
    """
    ctx.add_requirement(tc, 'require_admin', True)
    return tc


def require_command(command):
    """
    Disable test if given command is not provided
    """
    def wrapped(tc):
        if which(command) is None:
            tc.enabled = False
        return tc
    return wrapped



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020, Intel Corporation
#

import testframework as t


class Pmem2Memmove(t.Test):
    test_type = t.Short
    filesize = 4 * t.MiB
    envs0 = ()
    envs1 = ()
    test_cases = [
        # No offset, no overlap
        ['b:4096'],

        # aligned dest, unaligned source, no overlap
        ['s:7', 'b:4096'],

        # unaligned dest, unaligned source, no overlap
        ['d:7', 's:13', 'b:4096'],

        # all aligned, src overlaps dest
        ['b:4096', 's:23', 'o:1'],

        # unaligned destination
        ['b:4096', 'd:21'],

        # unaligned source and dest
        ['b:4096', 'd:21', 's:7'],

        # overlap of src, aligned src and dest
        ['b:4096', 'o:1', 's:20'],

        # overlap of src, aligned src, unaligned dest
        ['b:4096', 'd:13', 'o:1', 's:20'],

        # dest overlaps src, unaligned dest, aligned src
        ['b:2048', 'd:33', 'o:1'],

        # dest overlaps src, aligned dest and src
        ['b:4096', 'o:1', 'd:20'],

        # aligned dest, no overlap, small length
        ['b:8'],

        # small length, offset 1 byte from 64 byte boundary
        ['b:4', 'd:63'],

        # overlap, src < dest, small length (ensures a copy backwards,
        # with number of bytes to align < length)
        ['o:1', 'd:2', 'b:8']
    ]

    def run(self, ctx):
        for env in self.envs0:
            ctx.env[env] = '0'
        for env in self.envs1:
            ctx.env[env] = '1'
        for tc in self.test_cases:
            filepath = ctx.create_holey_file(self.filesize, 'testfile',)
            ctx.exec('pmem2_memmove', filepath, *tc)


class TEST0(Pmem2Memmove):
    pass


@t.require_architectures('x86_64')
class TEST1(Pmem2Memmove):
    envs0 = ("PMEM_AVX512F",)


@t.require_architectures('x86_64')
class TEST2(Pmem2Memmove):
    envs0 = ("PMEM_AVX512F", "PMEM_AVX",)


class TEST3(Pmem2Memmove):
    envs1 = ("PMEM_NO_MOVNT",)


class TEST4(Pmem2Memmove):
    envs1 = ("PMEM_NO_MOVNT", "PMEM_NO_GENERIC_MEMCPY")



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#
"""Parser for user provided test configuration"""

import argparse
import os
import string
import sys
from datetime import timedelta

import builds
import context as ctx
import granularity
import futils
import test_types
import valgrind as vg

try:
    import testconfig
except ImportError:
    sys.exit('Please add valid testconfig.py file - see testconfig.py.example')


class _ConfigFromDict:
    """
    Class fields are created from provided dictionary. Used for creating
    a final config object.
    """
    def __init__(self, dict_):
        for k, v in dict_.items():
            setattr(self, k, v)

    # special method triggered if class attribute was not found
    # https://docs.python.org/3.5/reference/datamodel.html#object.__getattr__
    def __getattr__(self, name):
        if name == 'page_fs_dir':
            raise futils.Skip('Configuration field "{}" not found. '
                              'No page granularity test directory '
                              'provided'.format(name))

        if name == 'cacheline_fs_dir':
            raise futils.Skip('Configuration field "{}" not found. '
                              'No cache line granularity test '
                              'directory provided'.format(name))

        if name == 'byte_fs_dir':
            raise futils.Skip('Configuration field "{}" not found. '
                              'No byte granularity test directory '
                              'provided'.format(name))

        raise AttributeError('Provided test configuration may be '
                             'invalid. No "{}" field found in '
                             'configuration.'
                             .format(name))


def _str2list(config):
    """
    Convert the string with test sequence to equivalent list.
    example:
    _str2list("0-3,6") -->  [0, 1, 2, 3, 6]
    _str2list("1,3-5") -->  [1, 3, 4, 5]
    """
    arg = config['test_sequence']
    if not arg:
        # test_sequence not set, do nothing
        return

    seq = []
    try:
        for number in arg.split(','):
            if '-' in number:
                number = number.split('-')
                begin = int(number[0])
                end = int(number[1])
                step = 1 if begin < end else -1
                seq.extend(range(begin, end + step, step))
            else:
                seq.append(int(number))

    except (ValueError, IndexError):
        print('Provided test sequence "{}" is invalid'.format(arg))
        raise

    config['test_sequence'] = seq


def _str2time(config):
    """
    Convert the string with s, m, h, d suffixes to time format

    example:
    _str2time("5s")  -->  "0:00:05"
    _str2time("15m") -->  "0:15:00"
    """
    string_ = config['timeout']
    try:
        timeout = int(string_[:-1])
    except ValueError as e:
        raise ValueError("invalid timeout argument: {}".format(string_)) from e
    else:
        if "d" in string_:
            timeout = timedelta(days=timeout)
        elif "m" in string_:
            timeout = timedelta(minutes=timeout)
        elif "h" in string_:
            timeout = timedelta(hours=timeout)
        elif "s" in string_:
            timeout = timedelta(seconds=timeout)

        config['timeout'] = timeout.total_seconds()


def _str2ctx(config):
    """Convert context classes from strings to actual classes"""
    def class_from_string(name, base):
        if name == 'all':
            return base.__subclasses__()

        try:
            return next(b for b in base.__subclasses__()
                        if str(b) == name.lower())
        except StopIteration:
            print('Invalid config value: "{}".'.format(name))
            raise

    def convert_internal(key, base):
        if not isinstance(config[key], list):
            config[key] = ctx.expand(class_from_string(config[key], base))
        else:
            classes = [class_from_string(cl, base) for cl in config[key]]
            config[key] = ctx.expand(*classes)

    convert_internal('build', builds.Build)
    convert_internal('test_type', test_types._TestType)
    convert_internal('granularity', granularity.Granularity)

    if config['force_enable'] is not None and config['force_enable'] != 'none':
        config['force_enable'] = next(
            t for t in vg.TOOLS
            if t.name.lower() == config['force_enable'])


class Configurator():
    """Parser for user test configuration.

    Configuration is generated from two sources: testconfig.py
    file and main script (RUNTESTS.py or TESTS.py files)
    command line arguments.
    Since these sources cannot change during script execution,
    the configurator class can be initialized multiple times
    throughout the implementation and always returns the same
    configuration result.

    Values from testconfig.py are overridden by respective
    values from command line arguments - provided the latter occur.

    Attributes:
        config: final test configuration meant to be used by
            the user
    """

    def __init__(self):
        self.config = self.parse_config()

    def parse_config(self):
        """
        Parse and return test execution config object. Final config is
        composed from 2 config values - values from testconfig.py file
        and values provided by command line args.
        """
        self.argparser = self._init_argparser()
        try:
            args_config = self._get_args_config()

            # The order of configs addition in 'config' initialization
            # is relevant - values from each next added config overwrite
            # values of already existing keys.
            config = {**testconfig.config, **args_config}

            self._convert_to_usable_types(config)

            # Remake dict into class object for convenient fields acquisition
            config = _ConfigFromDict(config)

            # device_dax_path may be either a single string with path
            # or a sequence of paths
            config.device_dax_path = futils.to_list(config.device_dax_path,
                                                    str)

            return config

        except KeyError as e:
            print("No config field '{}' found. "
                  "testconfig.py file may be invalid.".format(e.args[0]))
            raise

    def _convert_to_usable_types(self, config):
        """
        Converts config values types as parsed from user input into
        types usable by framework implementation
        """
        _str2ctx(config)
        _str2list(config)
        _str2time(config)

    def _get_args_config(self):
        """Return config values parsed from command line arguments"""

        # 'group' positional argument added only if RUNTESTS.py is the
        # execution entry point
        from_runtest = os.path.basename(sys.argv[0]) == 'RUNTESTS.py'
        if from_runtest:
            self.argparser.add_argument('group', nargs='*',
                                        help='Run only tests '
                                        'from selected groups')

        # remove possible whitespace and empty args
        sys.argv = [arg for arg in sys.argv if arg and not arg.isspace()]

        args = self.argparser.parse_args()

        if from_runtest:
            # test_sequence does not make sense if group is not set
            if args.test_sequence and not args.group:
                self.argparser.error('"--test_sequence" argument needs '
                                     'to have "group" arg set')

            # remove possible path characters added by shell hint
            args.group = [g.strip(string.punctuation) for g in args.group]

        # make into dict for type consistency
        return {k: v for k, v in vars(args).items() if v is not None}

    def _init_argparser(self):
        def ctx_choices(cls):
            return [str(c) for c in cls.__subclasses__()]

        parser = argparse.ArgumentParser()
        parser.add_argument('--fs_dir_force_pmem', type=int,
                            help='set PMEM_IS_PMEM_FORCE for tests run on'
                            ' pmem fs')
        parser.add_argument('-l', '--unittest_log_level', type=int,
                            help='set log level. 0 - silent, 1 - normal, '
                            '2 - verbose')
        parser.add_argument('--keep_going', type=bool,
                            help='continue execution despite test fails')
        parser.add_argument('-b', dest='build',
                            help='run only specified build type',
                            choices=ctx_choices(builds.Build), nargs='*')
        parser.add_argument('-g', dest='granularity',
                            choices=ctx_choices(granularity.Granularity),
                            nargs='*', help='run tests on a filesystem'
                            ' with specified granularity types.')
        parser.add_argument('-t', dest='test_type',
                            help='run only specified test type where '
                            'check = short + medium',
                            choices=ctx_choices(test_types._TestType),
                            nargs='*')
        parser.add_argument('-o', dest='timeout',
                            help="set timeout for test execution timeout: "
                            "integer with an optional suffix:''s' for seconds,"
                            " 'm' for minutes, 'h' for hours or 'd' for days.")
        parser.add_argument('-u', dest='test_sequence',
                            help='run only tests from specified test sequence '
                            'e.g.: 0-2,5 will execute TEST0, '
                            'TEST1, TEST2 and TEST5',
                            default='')
        parser.add_argument('--list-testcases', dest='list_testcases',
                            action='store_const', const=True,
                            help='List testcases only')
        parser.add_argument('--fail-on-skip', dest='fail_on_skip',
                            action='store_const', const=True,
                            help='Skipping tests also fail')
        parser.add_argument('--test-label', dest='test_label', default='',
                            action='store',
                            help='Run only tests marked with the label')

        tracers = parser.add_mutually_exclusive_group()
        tracers.add_argument('--tracer', dest='tracer', help='run C binary '
                             'with provided tracer command. With this option '
                             'stdout and stderr are not redirected, enabling '
                             'interactive sessions.',
                             default='')
        tracers.add_argument('--gdb', dest='tracer', action='store_const',
                             const='gdb --args', help='run gdb as a tracer')
        tracers.add_argument('--cgdb', dest='tracer', action='store_const',
                             const='cgdb --args', help='run cgdb as a tracer')

        fe_choices = [str(t) for t in vg.TOOLS] + ['none']
        parser.add_argument('--force-enable', choices=fe_choices, default=None,
                            help='allows to force the use of a specific '
                            'valgrind tool, but skips tests where the tool is '
                            'explicitly disabled')

        return parser



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2021-2023, Intel Corporation
#
import logging
import os
import shlex
import subprocess as sp
import tempfile
from os import path
import futils

DEFAULT_START_COMMANDS = ['set disable-randomization off',
                          'set width 0',
                          'set height 0',
                          'set verbose off',
                          'set confirm off']


class GdbProcess:
    """Class for invoking gdb from python program

       It writes and executes commands from a temporary gdb command file.

       Attributes:
           _process (CompletedProcess): finished process returned from run().
           _env (dict): environment variables

    """
    def __init__(self, env, cwd, testnum, path_to_exec, args_to_exec='',
                 gdb_options='', timeout=15):
        """
        Args:
            cwd (string): path to the test directory.
            testnum (string): test number.
            path_to_exec (string): path to executable for gdb subprocess.
            args_to_exec (string): arguments to executable, divided by spaces.
                Defaults to empty string.
            gdb_options (string): options for gdb, divided by spaces.
                Defaults to empty string.
            timeout (integer): timeout in seconds. Defaults to 15.
            commands (list): list of commands to gdb command file.
                Always contains default commands at its beginning.
            log_file (str): path to the gdb output log file.

        """
        self._process = None
        self._env = env
        self.cwd = cwd
        self.path_to_exec = path_to_exec
        self.args_to_exec = args_to_exec
        self.gdb_options = gdb_options
        self.timeout = timeout
        self.commands = []

        for command in DEFAULT_START_COMMANDS:
            self.add_command(command)

        log_name = 'python_gdb{}.log'.format(testnum)
        self.log_file = path.join(cwd, log_name)

        logging.basicConfig(filename=self.log_file, level=logging.INFO,
                            format='%(levelname)s:%(message)s')

    def execute(self):
        tmp = self._env.copy()
        futils.add_env_common(tmp, os.environ.copy())

        self.file_gdb = tempfile.NamedTemporaryFile(mode='r+')
        self.file_gdb.writelines('%s' % command for command in self.commands)
        self.file_gdb.seek(0)
        run_list = self._prepare_args()

        try:
            self._process = sp.run(args=run_list,
                                   env=tmp,
                                   stdin=sp.PIPE,
                                   stdout=sp.PIPE,
                                   stderr=sp.PIPE,
                                   universal_newlines=True,
                                   timeout=self.timeout)
        except sp.TimeoutExpired:
            self._close()
            raise

        logging.info(self._process.stdout)
        logging.error(self._process.stderr)

        self.validate_gdb()
        self._close()

    def add_command(self, command):
        if not command.endswith('\n'):
            command = ''.join([command, '\n'])
        self.commands.append(command)

    def validate_gdb(self):
        """Check if there was an error in gdb process"""
        if path.isfile(self.log_file + '.match'):
            # if there is a gdb log match file, do nothing - log file
            # will be checked by 'match' tool
            return
        if self._process.stderr:
            print(self._process.stderr)
            raise futils.Fail('Gdb validation failed')

    def _prepare_args(self):
        gdb_command = ''.join(['--command=', self.file_gdb.name])
        args_str = ' '.join(['gdb', self.gdb_options,
                             gdb_command, self.file_gdb.name,
                             '--args', self.path_to_exec,
                             self.args_to_exec])
        args = shlex.split(args_str)
        return args

    def _close(self):
        for handler in logging.root.handlers[:]:
            handler.close
            logging.root.removeHandler(handler)
        self.file_gdb.close()



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2019-2023, Intel Corporation
#

from os import path

import testframework as t
from testframework import granularity as g


@g.require_granularity(g.CACHELINE)
@t.require_build('nondebug')
class Base(t.Test):
    test_type = t.Long
    seed = '12345'
    defrag = '1'

    def run(self, ctx):
        testfile = path.join(ctx.testdir, 'testfile')
        # this test is extremely long otherwise
        ctx.env['PMEM_NO_FLUSH'] = '1'
        ctx.exec('obj_fragmentation2',
                 testfile, ctx.workload(), self.seed, self.defrag)


# These tests last too long under drd/helgrind/memcheck/pmemcheck
# Exceptions: workloads no. 6 and 8 under memcheck/pmemcheck (run with TEST1)
@t.require_valgrind_disabled('drd', 'helgrind', 'memcheck', 'pmemcheck')
@t.add_params('workload', [0, 1, 2, 3, 4, 5, 7, 9])
class TEST0(Base):
    pass


@t.require_valgrind_disabled('drd', 'helgrind')
@t.add_params('workload', [6, 8])
class TEST1(Base):
    pass



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2022, Intel Corporation
#


import testframework as t


class Pmem2MovntCommon(t.Test):
    test_type = t.Short
    filesize = 4 * t.MiB
    filepath = None

    def create_file(self, ctx):
        self.filepath = ctx.create_holey_file(self.filesize, 'testfile',)


class Pmem2Movnt(Pmem2MovntCommon):
    threshold = None
    threshold_values = ['1024', '5', '-15']
    envs0 = ()

    def run(self, ctx):
        super().create_file(ctx)
        for env in self.envs0:
            ctx.env[env] = '0'

        ctx.exec('pmem2_movnt', self.filepath)
        for tv in self.threshold_values:
            ctx.env['PMEM_MOVNT_THRESHOLD'] = tv
            ctx.exec('pmem2_movnt', self.filepath)


class TEST0(Pmem2Movnt):
    pass


class TEST1(Pmem2MovntCommon):
    def run(self, ctx):
        super().create_file(ctx)
        ctx.env['PMEM_NO_MOVNT'] = '1'
        ctx.exec('pmem2_movnt', self.filepath)


class TEST2(Pmem2MovntCommon):
    def run(self, ctx):
        super().create_file(ctx)
        ctx.env['PMEM_NO_MOVNT'] = '1'
        ctx.env['PMEM_NO_GENERIC_MEMCPY'] = '1'
        ctx.exec('pmem2_movnt', self.filepath)


@t.require_architectures('x86_64')
class TEST3(Pmem2Movnt):
    envs0 = ("PMEM_AVX512F",)


@t.require_architectures('x86_64')
class TEST4(Pmem2Movnt):
    envs0 = ("PMEM_AVX512F", "PMEM_AVX",)


@t.require_architectures('x86_64')
class TEST5(Pmem2Movnt):
    envs1 = ("PMEM_MOVDIR64B",)



#!../env.py
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2024, Intel Corporation
#

import testframework as t
from testframework import granularity as g
import valgrind as vg


# These tests last too long under drd
# Exceptions: test no. 2
@t.require_valgrind_disabled('drd')
class ObjDefragAdvanced(t.Test):
    test_type = t.Short

    max_nodes = 50
    max_edges = 10
    graph_copies = 10
    pool_size = 500 * t.MiB
    max_rounds = 10
    min_root_size = 0

    def run(self, ctx):
        ctx.require_free_space(self.pool_size)

        path = ctx.create_holey_file(self.pool_size, 'testfile')
        dump1 = 'dump_1_{}.log'.format(self.testnum)
        dump2 = 'dump_2_{}.log'.format(self.testnum)

        ctx.exec('obj_defrag_advanced',
                 'op_pool_create', path,
                 'op_graph_create', str(self.max_nodes), str(self.max_edges),
                 str(self.graph_copies), str(self.min_root_size),
                 'op_graph_dump', dump1,
                 'op_graph_defrag', str(self.max_rounds),
                 'op_graph_dump', dump2,
                 'op_pool_close',
                 'op_dump_compare', dump1, dump2)


class TEST0(ObjDefragAdvanced):
    max_nodes = 5
    max_edges = 5
    graph_copies = 5


class TEST1(ObjDefragAdvanced):
    max_nodes = 2048
    max_edges = 5
    graph_copies = 5


@t.require_valgrind_disabled('helgrind')
@g.require_granularity(g.CACHELINE)
class TEST2(ObjDefragAdvanced):
    test_type = t.Medium
    # XXX port this to the new framework
    # Restore defaults
    drd = vg.AUTO

    max_nodes = 512
    max_edges = 64
    graph_copies = 5
    min_root_size = 4096


@g.require_granularity(g.CACHELINE)
class ObjDefragAdvancedMt(ObjDefragAdvanced):
    test_type = t.Medium

    nthreads = 2
    ncycles = 2

    def run(self, ctx):
        ctx.require_free_space(self.pool_size)

        path = ctx.create_holey_file(self.pool_size, 'testfile')

        ctx.exec('obj_defrag_advanced',
                 'op_pool_create', path,
                 'op_graph_create_n_defrag_mt', self.max_nodes,
                 self.max_edges, self.graph_copies, self.min_root_size,
                 self.max_rounds, self.nthreads, self.ncycles, self.testnum,
                 'op_pool_close')


class TEST3(ObjDefragAdvancedMt):
    test_type = t.Long

    max_nodes = 256
    max_edges = 64
    graph_copies = 10
    nthreads = 1
    ncycles = 25


class TEST4(ObjDefragAdvancedMt):
    max_nodes = 128
    max_edges = 32
    graph_copies = 10
    nthreads = 10
    ncycles = 25


# XXX disable the test for 'drd'
# until https://github.com/pmem/pmdk/issues/5690 is fixed.
# previousely the test has been disabled for other Valgrind options
# This test last too long under helgrind/memcheck/pmemcheck
# @t.require_valgrind_disabled('helgrind', 'memcheck', 'pmemcheck')
@t.require_valgrind_disabled('helgrind', 'memcheck', 'pmemcheck', 'drd')
class TEST5(ObjDefragAdvancedMt):

    max_nodes = 256
    max_edges = 32
    graph_copies = 5
    nthreads = 10
    ncycles = 25


# a testcase designed to verify the pool content in case of fail
# class TESTX(ObjDefragAdvanced):
#     def run(self, ctx):
#         path = '/custom/pool/path'
#         dump_prefix = 'dump'
#
#         ctx.exec('obj_defrag_advanced',
#             'op_pool_open', path,
#             'op_graph_dump_all', dump_prefix,
#             'op_pool_close')



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2018-2023, Intel Corporation

"""
This module includes functions to check if documentation of PMDK library
is built correctly and tests which use these functions.

Tests check:
-if all macros from header files are included in the doc directory,
-if all functions from .so files are included in the doc directory,
-if all functions and macros from the doc directory, are included
 in headers files and .so files,
-if all functions and macros have direct access to man page through use
 command 'man function_name/macro_name'.

Required argument:
  -r <PMDK_path>            the PMDK library root path.
  -i <PMDK_install_path>    PMDK library installation path.
"""

import unittest
from subprocess import check_output, call, DEVNULL
from os import listdir, path, linesep, walk
import re
import sys

# definitions of regexes
EXPRESSION_AT_THE_LINE_START = r'[\s]*([a-zA-Z_]+)[\\(\s]+'
EXPRESSION_AFTER_DEFINE_PHRASE = r'[\s]*#define[\s]*([a-zA-Z_]+)[\\(\s]+'

PMDK_LIBRARIES = ['libpmem', 'libpmemobj', 'libpmempool', 'libpmem2']


def get_exceptions(pmdk_path):
    """
    Returns exceptions in a list. File 'exceptions.txt' contains all macros,
    which do not have a description in a man page by design.
    """
    exceptions_file_path =\
        path.join(path.dirname(path.realpath(__file__)), 'exceptions.txt')
    with open(exceptions_file_path, 'r') as f:
        exceptions = [line.strip().lower() for line in f]
    return exceptions


def parse_macro_name(exceptions, line, const_expression, macros):
    """
    Parses the name of the macro and adds it to existing list.
    """
    macro_name = re.search(const_expression, line).group(1).lower()
    if macro_name not in exceptions:
        macros.append(macro_name)


def get_macros(pmdk_path):
    """
    Parses all headers from the libpmemobj library, extracts all macros, and
    returns them in a list.
    """
    macros = []
    exceptions = get_exceptions(pmdk_path)
    is_macro_in_next_line = False
    includes_path = path.join(pmdk_path, 'src', 'include', 'libpmemobj')
    # Definition of macro occurs after the phrases: 'static inline' or '#define'
    for header_file in listdir(includes_path):
        with open(path.join(includes_path, header_file), 'r') as f:
            file_lines = f.readlines()
        for line in file_lines:
            if line.startswith('static inline'):
                is_macro_in_next_line = True
            elif is_macro_in_next_line:
                parse_macro_name(exceptions, line,
                                 EXPRESSION_AT_THE_LINE_START, macros)
                is_macro_in_next_line = False
            elif '#define' in line:
                parse_macro_name(exceptions, line,
                                 EXPRESSION_AFTER_DEFINE_PHRASE, macros)
    return macros


def get_functions_from_so_files(lib_path):
    """
    Returns names of functions in a list based on symbols 'T' from .so files.
    """
    functions_from_so_files = []
    path_to_so_files = lib_path
    for elem in listdir(path_to_so_files):
        if elem.endswith('.so') and elem.split('.')[0] in PMDK_LIBRARIES:
            process = check_output(
                'nm ' + elem + ' | grep " T "', cwd=path_to_so_files,
                shell=True)
            out = process.decode('UTF-8')
            for line in out.split(linesep):
                if line:
                    name = line.split(' ')[2].strip()
                    # Exclude'_pobj_debug_notice', because it is a name of
                    # a function which is not described in the man page
                    # by design.
                    if name != '_pobj_debug_notice':
                        functions_from_so_files.append(name)
    return functions_from_so_files


def get_functions_and_macros_from_doc(pmdk_path):
    """
    Returns names of functions and macros in a list based on names of files
    from the doc directory.
    """
    path_to_functions_and_macros = path.join(pmdk_path, 'doc')
    functions_and_macros_from_doc = []
    for _, _, files in walk(path_to_functions_and_macros):
        for f in files:
            if not f.endswith('.3'):
                continue
            # Files with extension '.3' have the same name as functions and
            # macros of PMDK library. 'pmemobj_action' is excluded, because
            # it is not a name of the function.
            if f.startswith('pmemobj_action'):
                continue
            functions_and_macros_from_doc.append(f.split('.')[0])
    return functions_and_macros_from_doc


def check_completeness_of_macros_in_doc(pmdk_path):
    """
    Returns missing macros in the doc directory in a list, based on macros
    extracted from header files.
    """
    macros = get_macros(pmdk_path)
    functions_from_doc = get_functions_and_macros_from_doc(pmdk_path)
    missing_macros = [macro for macro in macros if macro not in
                      functions_from_doc]
    return missing_macros


def check_completeness_of_functions_in_doc(pmdk_path, lib_path):
    """
    Returns list of missing macros in the doc directory, based on macros
    extracted from .so files.
    """
    functions_from_so_files = get_functions_from_so_files(lib_path)
    functions_from_doc = get_functions_and_macros_from_doc(pmdk_path)
    missing_functions_in_doc = [function for function in functions_from_so_files
                                if function not in functions_from_doc]
    return missing_functions_in_doc


def check_completeness_of_extracted_functions_and_macros(pmdk_path, lib_path):
    """
    Returns list of missing macros in header files and missing functions in .so
    files, based on functions and macros from the doc directory.
    """
    functions_from_doc = get_functions_and_macros_from_doc(pmdk_path)
    macros = get_macros(pmdk_path)
    functions_from_so_files = get_functions_from_so_files(lib_path)
    missing_functions_and_macros_in_doc = [
        item for item in functions_from_doc
        if (item not in macros and item not in functions_from_so_files)]
    return missing_functions_and_macros_in_doc


def check_linking_manpages(pmdk_path, doc_path):
    """
    Checks if macros and functions from the doc directory have
    direct access to man page through command 'man function_name/macro_name'
    and returns list with elements which are not linked with the man page.
    """
    functions_from_doc = get_functions_and_macros_from_doc(pmdk_path)
    functions_without_manpage = []
    for function in functions_from_doc:
        if doc_path:
            process_find_in_manpage = call(
                'man -M ' + doc_path + ' ' + function, shell=True,
                stdout=DEVNULL, stderr=DEVNULL)
        else:
            process_find_in_manpage = call(
                'man ' + function, shell=True, stdout=DEVNULL, stderr=DEVNULL)
        if process_find_in_manpage:
            functions_without_manpage.append(function)
    return functions_without_manpage


def parse_argument(argument_option):
    """
    Parses an option from the command line.
    """
    index = sys.argv.index(argument_option)
    try:
        argument = sys.argv[index+1]
    except IndexError:
        print('ERROR: Invalid argument!')
        print(__doc__)
        print(unittest.main.__doc__)
    else:
        sys.argv.pop(index)
        sys.argv.pop(index)
        return argument


class TestDocumentation(unittest.TestCase):

    def test_completeness_of_macros_in_doc(self):
        """
        Checks if all extracted macros from header files are in the doc
        directory.
        """
        missing_macros = check_completeness_of_macros_in_doc(pmdk_path)
        error_msg = linesep + 'List of missing macros in the doc directory:'
        for macro in missing_macros:
            error_msg += linesep + macro
        self.assertFalse(missing_macros, error_msg)

    def test_completeness_of_functions_in_doc(self):
        """
        Checks if all extracted functions from .so files are in the doc
        directory.
        """
        missing_functions =\
            check_completeness_of_functions_in_doc(pmdk_path, lib_path)
        error_msg = linesep + 'List of missing functions in the doc directory:'
        for function in missing_functions:
            error_msg += linesep + function
        self.assertFalse(missing_functions, error_msg)

    def test_completeness_of_extracted_functions_and_macros(self):
        """
        Checks if all functions and macros from the doc directory,
        are extracted from header files and .so files.
        """
        missing_functions_and_macros =\
            check_completeness_of_extracted_functions_and_macros(
                pmdk_path, lib_path)
        error_msg = linesep + 'List of missing macros and functions:'
        for elem in missing_functions_and_macros:
            error_msg += linesep + elem
        self.assertFalse(missing_functions_and_macros, error_msg)

    def test_linking_manpage_for_functions_and_macros(self):
        """
        Checks if all functions and macros have direct access to the man page
        through use command 'man function_name/macro_name'.
        """
        no_linked_functions_and_macros =\
            check_linking_manpages(pmdk_path, doc_path)
        error_msg =\
            linesep + 'List of macros and functions without the linked man\
            page:'
        for elem in no_linked_functions_and_macros:
            error_msg += linesep + elem
        self.assertFalse(no_linked_functions_and_macros, error_msg)


if __name__ == '__main__':
    doc_path = ''
    lib_path = ''
    if '-h' in sys.argv or '--help' in sys.argv:
        print(__doc__)
        unittest.main()
    elif '-r' in sys.argv:
        pmdk_path = parse_argument('-r')
        if '-i' in sys.argv:
            installation_path = parse_argument('-i')
            doc_path = path.join(installation_path, 'share', 'man')
            lib_path = path.join(installation_path, 'lib')
        else:
            print(__doc__)
            print(unittest.main.__doc__)
            exit(1)
        unittest.main()
    else:
        print(__doc__)
        print(unittest.main.__doc__)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2021-2023, Intel Corporation
#
"""Bad block utilities"""

import abc

import futils
import tools


class BBTool(abc.ABC):
    """
    An abstract class implementing an interface necessary to serve as an
    bad block handler tool for public BadBlock class.
    """

    @abc.abstractmethod
    def inject_bad_block(self, file, offset):
        """
        Inject an error into the namespace of the specified file at the
        provided location.
        """
        pass

    @abc.abstractmethod
    def get_bad_blocks_count(self, file):
        """
        Returns the number of bad blocks in the namespace of
        the specified file.
        """
        pass

    @abc.abstractmethod
    def clear_all_bad_blocks(self, file):
        """
        Clears all bad blocks of a file.
        """
        pass


class BadBlock:
    """
    Utility class specifying high-level workflow for handling bad block.
    """
    def __init__(self, util_tools, tool: BBTool = None):
        if tool:
            if not isinstance(tool, BBTool):
                raise futils.Fail('{} should be a subclass of BBTool'
                                  .format(tool))
            self.tool = tool
        else:
            self.tool = NdctlBB(util_tools)

    def inject(self, file, offset):
        self.tool.inject_bad_block(file, offset)

    def get_count(self, file):
        return self.tool.get_bad_blocks_count(file)

    def clear_all(self, file):
        self.tool.clear_all_bad_blocks(file)


class NdctlBB(tools.Ndctl, BBTool):
    """ndctl based bad-block tool"""

    def __init__(self, util_tools):
        super().__init__()
        self.util_tools = util_tools

    def inject_bad_block(self, file, offset):
        device = self._get_path_device(file)
        namespace = self.get_dev_namespace(device)

        count = 1

        if self.is_devdax(device):
            # for testing purposes devdax is treated as a file itself,
            # therefore the offset is a physical offset on the device
            block = offset
        else:
            # fsdax relies on the fs for block allocation, so a virtual
            # offset must be translated into a physical one with a tool
            proc = self.util_tools.extents("-l {}".format(offset), file)
            if proc.returncode != 0:
                raise futils.Fail('unable to translate bad blocks')
            block = int(proc.stdout.strip())

        futils.run_command("sudo ndctl inject-error --block={} --count={} {}"
                           .format(block, count, namespace),
                           "injecting bad block failed")

    def get_bad_blocks_count(self, file):
        device = self._get_path_device(file)
        return self.get_dev_bb_count(device)

    def clear_all_bad_blocks(self, file):
        device = self._get_path_device(file)
        namespace = self.get_dev_namespace(device)

        if self.is_devdax(device):
            futils.run_command("sudo ndctl clear-errors {}".format(namespace),
                               "failed to clear bad blocks")
            # The "sudo ndctl clear-errors" command resets permissions
            # of the DAX block device and its resource files
            # because it removes the old files and creates new ones
            # with default permissions which interrupts further test execution.
            # The following commands set permissions which should allow
            # further test execution to work around this issue.
            futils.run_command("sudo chmod a+rw {}".format(device),
                               "failed to change permissions of a DAX device")
            resources = "/sys/bus/nd/devices/ndbus*/region*/dax*/resource"
            futils.run_command("sudo chmod a+r {}".format(resources),
                               "failed to change permissions "
                               "of DAX devices' resources")
        else:
            out = futils.run_command("mount | grep {}".format(device),
                                     "querying mount points failed")
            mount = out.strip().decode('UTF8').split()[2]

            futils.run_command("sudo umount {}".format(mount),
                               "unmounting device failed")

            futils.run_command("sudo ndctl clear-errors {}".format(namespace),
                               "clearing bad blocks failed")

            futils.run_command("sudo mount -o dax {} {}".format(device, mount),
                               "mounting device failed")



#!../env.py
#
# SPDX-License-Identifier: BSD-3-Clause
# Copyright 2020-2023, Intel Corporation

import testframework as t
from testframework import granularity as g


@g.require_granularity(g.ANY)
class PMEM2_SOURCE(t.Test):
    test_type = t.Short

    def run(self, ctx):
        filepath = ctx.create_holey_file(16 * t.MiB, 'testfile1')
        ctx.exec('pmem2_source', self.test_case, filepath)


@g.no_testdir()
class PMEM2_SOURCE_NO_DIR(t.Test):
    test_type = t.Short

    def run(self, ctx):
        ctx.exec('pmem2_source', self.test_case)


@t.require_devdax(t.DevDax('devdax1'))
class PMEM2_SOURCE_DEVDAX(t.Test):
    test_type = t.Short

    def run(self, ctx):
        ddpath = ctx.devdaxes.devdax1.path
        ctx.exec('pmem2_source', self.test_case, ddpath)


class TEST0(PMEM2_SOURCE):
    """setting a read + write file descriptor in pmem2_source"""
    test_case = "test_set_rw_fd"


class TEST1(PMEM2_SOURCE):
    """setting a read only file descriptor in pmem2_source"""
    test_case = "test_set_ro_fd"


class TEST2(PMEM2_SOURCE):
    """setting invalid (closed) file descriptor in pmem2_source"""
    test_case = "test_set_invalid_fd"


class TEST3(PMEM2_SOURCE):
    """setting a write only file descriptor in pmem2_source"""
    test_case = "test_set_wronly_fd"


class TEST4(PMEM2_SOURCE):
    """allocation of pmem2_source in case of missing memory in system"""
    labels = ['fault_injection']
    test_case = "test_alloc_src_enomem"


class TEST5(PMEM2_SOURCE_NO_DIR):
    """deleting null pmem2_source"""
    test_case = "test_delete_null_config"


class TEST11(PMEM2_SOURCE):
    """set directory's fd in the source"""
    test_case = "test_set_directory_fd"

    def run(self, ctx):
        ctx.exec('pmem2_source', self.test_case, ctx.testdir)


class TEST13(PMEM2_SOURCE):
    """get file descriptor from the source"""
    test_case = "test_get_fd"


class TEST15(PMEM2_SOURCE_NO_DIR):
    """get file descriptor from the invalid source type"""
    test_case = "test_get_fd_inval_type"


class TEST16(PMEM2_SOURCE):
    """test mcsafe read operation"""
    test_case = "test_pmem2_src_mcsafe_read"


class TEST17(PMEM2_SOURCE):
    """test mcsafe write operation"""
    test_case = "test_pmem2_src_mcsafe_write"


class TEST18(PMEM2_SOURCE_DEVDAX):
    """test mcsafe read operation"""
    test_case = "test_pmem2_src_mcsafe_read"


class TEST19(PMEM2_SOURCE_DEVDAX):
    """test mcsafe write operation"""
    test_case = "test_pmem2_src_mcsafe_write"


class TEST20(PMEM2_SOURCE):
    """
    test mcsafe read and write operations with length bigger than source size
    """
    test_case = "test_pmem2_src_mcsafe_read_write_len_out_of_range"


class TEST21(PMEM2_SOURCE_DEVDAX):
    """
    test mcsafe read and write operations with length bigger than source size
    on devdax
    """
    test_case = "test_pmem2_src_mcsafe_read_write_len_out_of_range"


class TEST22(PMEM2_SOURCE):
    """
    test mcsafe read and write operations on source with invalid type
    """
    test_case = "test_pmem2_src_mcsafe_read_write_invalid_ftype"


class TEST23(PMEM2_SOURCE_DEVDAX):
    """
    test mcsafe read and write operations on source with invalid type on devdax
    """
    test_case = "test_pmem2_src_mcsafe_read_write_invalid_ftype"
